--[[
    TotemManager
    Copyright (C) 2008 Gael Lalire
    Contact: gael.lalire@laposte.net

    TotemManager is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    TotemManager is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with self.  If not, see <http://www.gnu.org/licenses/>.
--]]

-- begin import
local factoryLib = LibStub("FactoryLib-1.0");
local Factory = factoryLib.Factory;
local environmentLib = LibStub("EnvironmentLib-1.0");
local environment = environmentLib.environment;

local circleActionBarRendererLib = LibStub("CircleActionBarRendererLib-1.0");
local squareActionBarRendererLib = LibStub("SquareActionBarRendererLib-1.0");

local timerLib = LibStub("TimerLib-1.0");
local timerUpdaterLib = LibStub("TimerUpdaterLib-1.0");

local actionLib = LibStub("ActionLib-1.0");
local actionBarLib = LibStub("ActionBarLib-1.0");

local ActionBarRenderer = actionBarLib.ActionBarRenderer;
local Action = actionLib.Action;

local managerFrameLib = LibStub("ManagerFrameLib-1.0");
-- end import


local TotemManagerFrameHandler, TotemManagerFrameHandlerFactory, totemManagerFrameHandlerFactory = factoryLib.Factory.CreateType();

function TotemManagerFrameHandler:GetDirection()
  return self.frame:GetDirection();
end

function TotemManagerFrameHandler:SetDirection(_, direction)
  if self.totemManager:CheckCombat() then
    return;
  end
  self.frame:SetDirection(tonumber(direction));
end

function TotemManagerFrameHandler:IsVisible()
  return self.frame:IsVisible();
end

function TotemManagerFrameHandler:ToggleVisibility()
  if self.totemManager:CheckCombat() then
    return;
  end
  self.frame:ToggleVisibility();
end

function TotemManagerFrameHandler:GetAngle()
  return self.frame:GetAngle()*180/math.pi;
end

function TotemManagerFrameHandler:SetAngle(_, value)
  if self.totemManager:CheckCombat() then
    return;
  end
  
  local nvalue = tonumber(value);
  if nvalue>-90 and nvalue<90 then
    self.frame:SetAngle(value*math.pi/180);
  end
end

function TotemManagerFrameHandler:IsLibSinkMessageVisible()
  return self.frame:IsLibSinkMessageVisible();
end

function TotemManagerFrameHandler:ToggleLibSinkMessageVisibility()
  self.frame:ToggleLibSinkMessageVisibility();
end


function TotemManagerFrameHandlerFactory:Create(totemManager, frame)
  local totemManagerFrame = self:NewInstance();
  totemManagerFrame.totemManager = totemManager;
  totemManagerFrame.frame = frame;
  return totemManagerFrame;
end

-- Ankh cooldown updater (only shaman)
local AnkhTimerUpdater, ankhTimerUpdater = factoryLib.Factory.CreateSingletonType(timerUpdaterLib.TimerUpdaterFactory, {"UNIT_INVENTORY_CHANGED"});

function AnkhTimerUpdater:Enable(timer)
  timerUpdaterLib.TimerUpdater.Enable(self, timer);
  self:UNIT_INVENTORY_CHANGED(timer);
end

function AnkhTimerUpdater:UNIT_INVENTORY_CHANGED(timer)
  timer:SetCount(GetItemCount(self.ankh));  
end

TotemManager, totemManager = factoryLib.Factory.CreateSingletonType();

totemManagerData = {};

function TotemManager:BindButton(button, shortcut1, shortcut2, mouseClick)
  if button then
    ClearOverrideBindings(button);
    if shortcut1 then
      SetOverrideBindingClick(button, false, shortcut1, button:GetName(), mouseClick);
    end
    if shortcut2 then
      SetOverrideBindingClick(button, false, shortcut2, button:GetName(), mouseClick);
    end
  end
end

function TotemManager:AddTotemBind(totemName, totemSlot, bindingName)
  local shortcut1, shortcut2 = GetBindingKey(bindingName);
  self:BindButton(getglobal(totemName.."TotemSelected"), shortcut1, shortcut2);
end


function TotemManager:UPDATE_BINDINGS()
  if InCombatLockdown() then
    return;
  end

  local shortcut1, shortcut2;
  
  self:AddTotemBind("air", 4, "TOTEM_MANAGER_AIR");
  self:AddTotemBind("water", 3, "TOTEM_MANAGER_WATER");
  self:AddTotemBind("fire", 1, "TOTEM_MANAGER_FIRE");
  self:AddTotemBind("earth", 2, "TOTEM_MANAGER_EARTH");
  
  shortcut1, shortcut2 = GetBindingKey("TOTEM_MANAGER_SELECTED_CALL");
  self:BindButton(getglobal("totemCallSelected"), shortcut1, shortcut2);
  
  shortcut1, shortcut2 = GetBindingKey("TOTEM_MANAGER_EARTH_SHIELD");
  self:BindButton(getglobal("earthShieldSelected"), shortcut1, shortcut2);

  shortcut1, shortcut2 = GetBindingKey("TOTEM_MANAGER_MAIN_HAND_WEAPON");
  self:BindButton(getglobal("mainHandWeaponSelected"), shortcut1, shortcut2);

  shortcut1, shortcut2 = GetBindingKey("TOTEM_MANAGER_SHIELD");
  self:BindButton(getglobal("shieldSelected"), shortcut1, shortcut2);

  shortcut1, shortcut2 = GetBindingKey("TOTEM_MANAGER_TOTEMIC_CALL");
  self:BindButton(getglobal("totemicCall"), shortcut1, shortcut2);

end

function TotemManager:CheckCombat()
  if InCombatLockdown() then
    self.console:Print("You are in combat");
    return true;
  else
    return false;
  end
end

function TotemManager:ToggleResetAsked()
  if totemManagerData.alreadyActivate4_0_0 then
    totemManagerData.alreadyActivate4_0_0 = false;
    self.console:Print("Configuration will be reset, you need to restart your UI");
  else
    totemManagerData.alreadyActivate4_0_0 = true;
    self.console:Print("Reset avoided");
  end
end

function TotemManager:IsResetAsked()
  return not totemManagerData.alreadyActivate4_0_0;
end

function TotemManager:SetTooltipVisibility(_, value)
  ActionBarRenderer.tooltipVisibility = tonumber(value);
  totemManagerData.tooltipVisibility = value;
end

function TotemManager:GetTooltipVisibility()
  if ActionBarRenderer.tooltipVisibility == nil then
    ActionBarRenderer.tooltipVisibility = 1;
  end
  return ActionBarRenderer.tooltipVisibility;
end

function TotemManager:IsBlizzFrameVisible()
  return totemManagerData.blizzFrameVisible;
end

function TotemManager:ToggleBlizzFrameVisibility()
  if self:CheckCombat() then
    return;
  end
  if totemManagerData.blizzFrameVisible then
    MultiCastActionBarFrame:Hide();
    totemManagerData.blizzFrameVisible = false;
  else
    MultiCastActionBarFrame:Show();
    totemManagerData.blizzFrameVisible = true;
  end
end

function TotemManager:IsLocked()
  return self.locked;
end

function TotemManager:ToggleLocked()
  if self:CheckCombat() then
    return;
  end
  self.locked = not self.locked;
  for _, frame in pairs(self.frames) do
    frame:ToggleLocked();
  end
end

function TotemManager:GetTimerSkin()
  if totemManagerData.timerSkin == nil then
    totemManagerData.timerSkin = 1;
  end
  return totemManagerData.timerSkin;
end

function TotemManager:SetTimerSkin(_, value)
  if self:CheckCombat() then
    return;
  end
  totemManagerData.timerSkin = tonumber(value);
end

function TotemManager:GetActionBarSkin()
  if totemManagerData.actionBarSkin == nil then
    -- circle by default
    totemManagerData.actionBarSkin = 2;
  end
  return totemManagerData.actionBarSkin;
end

function TotemManager:SetActionBarSkin(_, value)
  if self:CheckCombat() then
    return;
  end
  local nvalue = tonumber(value);
  local rendererCreatorFunc = self.actionBarSkin[nvalue];
  if rendererCreatorFunc ~= nil then
    for _, frame in pairs(self.frames) do
      frame:SetActionBarRendererCreatorFunc(rendererCreatorFunc);
    end
    totemManagerData.actionBarSkin = nvalue;
  end
end

function TotemManager:GetScale()
  if totemManagerData.scale == nil then
    totemManagerData.scale = 1;
  end
  return totemManagerData.scale;
end

function TotemManager:SetScale(_, value)
  if self:CheckCombat() then
    return;
  end
  local nvalue = tonumber(value);
  if nvalue>0 then
    totemManagerData.scale = nvalue;
    for _, frame in pairs(self.wowFrame) do  
      frame:SetScale(nvalue);
      frame:Raise();
    end
  end
end

function TotemManager:OpenGui()
  InterfaceOptionsFrame_OpenToCategory(self.optionFrame);
end

function TotemManager:CreateSetMacro()
  if self:CheckCombat() then
    return;
  end
  local setNum = 1;
  for _, macroName in ipairs({"RAID", "FARM", "BG", "ARN"}) do 
    local setNum2 = setNum + 1;
    local macroBody = "/click [modifier:shift, spec:1] totemManagerSet S"..setNum..";";
    macroBody = macroBody.."[modifier:shift, spec:2] totemManagerSet S"..setNum2..";";
    macroBody = macroBody.."[modifier:alt, spec:1] totemManagerSet D"..setNum..";";
    macroBody = macroBody.."[modifier:alt, spec:2] totemManagerSet D"..setNum2..";";
    macroBody = macroBody.."[spec:1] totemManagerSet R"..setNum..";";
    macroBody = macroBody.."[spec:2] totemManagerSet R"..setNum2..";";
    setNum = setNum + 2;
    CreateMacro(macroName, 1, macroBody, 1);
  end
end

function TotemManager:CreateConsole()

  local options = {
    type = "group",
    handler = self;
    args = {
      timerSkin = {
        type = "select",
        name = "timer skin",
        desc = "change timer skin (need to restart UI)",
        get = "GetTimerSkin",
        set = "SetTimerSkin",
        values = {"TM2"},
        order = 1,
      },
      actionBarSkin = {
        type = "select",
        name = "action bar skin",
        desc = "change action bar skin (need to restart UI)",
        get = "GetActionBarSkin",
        set = "SetActionBarSkin",
        values = {"TM2", "Circle"},
        order = 1,
      },
      lock = {
        type = "toggle",
        name = "locked",
        desc = "lock or unlock all frames",
        set = "ToggleLocked",
        get = "IsLocked"
      },
      --[[
      blizzFrame = {
        type = "toggle",
        name = "blizzFrame",
        desc = "show or hide blizz totem frames",
        set = "ToggleBlizzFrameVisibility",
        get = "IsBlizzFrameVisible"
      },
      ]]
      tooltip = {
        type = "select",
        name = "tooltip",
        desc = "tooltip visibility",
        get = "GetTooltipVisibility",
        set = "SetTooltipVisibility",
        values = {"Hide while in combat", "Show", "Hide"},
        order = 1,
      },
      scale = {
        type = "range",
        name = "scale",
        desc = "change frames scaling",
        min = .5,
        max = 2,
        step = .01,  
        get = "GetScale",
        set = "SetScale",
      },
      reset = {
        type = "toggle",
        name = "reset on next UI load",
        desc = "if the box is checked, default values will be loaded on next UI load",
        set = "ToggleResetAsked",
        get = "IsResetAsked",
      },
      createSetMacro = {
        type = "execute",
        name = "create set macro",
        desc = "create RAID, FARM, BG, ARN macro",
        func = "CreateSetMacro",
      },
    }    
  };
  
  for frameName, frame in pairs(self.frames) do
    options.args[frameName] = {type="group", name=frameName, handler=totemManagerFrameHandlerFactory:Create(self, frame), args = {
      visibility = {
        type = "toggle",
        name = "visible",
        desc = "show or hide this frame",
        set = "ToggleVisibility",
        get = "IsVisible"        
      },
      libSinkVisibility = {
        type = "toggle",
        name = "libSinkVisible",
        desc = "show or hide LibSink messages",
        set = "ToggleLibSinkMessageVisibility",
        get = "IsLibSinkMessageVisible"        
      },      
      mode = {
        type = "select",
        name = "mode",
        desc = "direction of manager",
        get = "GetDirection",
        set = "SetDirection",
        values = {"Left mode", "Right mode"},
        order = 1,
      },
      angle = {
        type = "range",
        name = "angle",
        desc = "change orientation",
        min = -90,
        max = 90,
        step = 0.1,  
        get = "GetAngle",
        set = "SetAngle",
      },
    }};
    
  end
  
  local console = {};
  self.console = console;
  setmetatable(console, {__index = LibStub("AceConsole-3.0"), __tostring = function ()
    return "TotemManager";
  end});
  
  LibStub("AceConfig-3.0"):RegisterOptionsTable("TotemManager", options);
  self.optionFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("TotemManager", "TotemManager");
  console:RegisterChatCommand("totemmanager", "ChatCommand");
  console:RegisterChatCommand("tm", "ChatCommand");

  function console:ChatCommand(input)
    if not input or input:trim() == "" then
      totemManager:OpenGui();
    else
      LibStub("AceConfigCmd-3.0").HandleCommand(totemManager, "tm", "TotemManager", input);
    end
  end  

end

function TotemManager:Activate()
  ActionBarRenderer.tooltipVisibility = totemManagerData.tooltipVisibility;
  
  ankhTimerUpdater.ankh = self.ANKH;
  
  self.locked = true;
  
  self.setButtons = {};
  
  -- hide default blizzard timer
  TotemFrame:UnregisterEvent("PLAYER_TOTEM_UPDATE");
  TotemFrame:UnregisterEvent("PLAYER_ENTERING_WORLD");  
  TotemFrame:SetScript("OnShow", function ()
    TotemFrame:Hide();
  end);
  TotemFrame:Hide();
  
  --[[
  if totemManagerData.blizzFrameVisible then
    MultiCastActionBarFrame:Show();
  else
    MultiCastActionBarFrame:Hide();  
  end
  ]]
    
  self.left = 300;
  self.top = 200;

  self.actionBarConfList = {};
  self.removeFuncMap = {};
  
  local updater = timerUpdaterLib.totemTimerUpdaterFactory:Create(4);

  self.actionBarConfList["airTotem"] = {
    {name=self.GROUNDING, effect=self.GROUNDING_EFFECT, duration = function (rank)
      local focRank = environment:GetTalentRankByIndex(5, 3);
      if focRank == 0 then
        return 45;
      elseif  focRank == 1 then
        return 63*5/6;
      else
        return 81*5/7;
      end
    end, updater=updater},
    {name=self.WINDFURY, raidEffect=true, effect=self.WINDFURY_EFFECT, updater=updater},
    {name=self.WRATH_AIR, raidEffect=true, effect=self.WRATH_AIR_EFFECT, updater=updater},
    
  }
  self.removeFuncMap["airTotem"] = function () DestroyTotem(4) end;
  table.insert(self.setButtons, "airTotemSet");
  
  updater = timerUpdaterLib.totemTimerUpdaterFactory:Create(3);

  self.actionBarConfList["waterTotem"] = {
    {name=self.FIRE_RESISTANCE, raidEffect=true, effect=self.FIRE_RESISTANCE_EFFECT, updater=updater},
    {name=self.HEALING_STREAM, tickFrequency = 2, updater=updater},
    {name=self.ENAMORED_WATER_SPIRIT, totemName=self.ANCIENT_MANA_SPRING, action="item", duration = 24, effect=self.ANCIENT_MANA_SPRING_EFFECT, updater=updater},
    {name=self.MANA_SPRING, raidEffect=true, tickFrequency = 2, updater=updater},
    {name=self.MANA_TIDE, duration = 12, tickFrequency = 3, updater=updater},
    {name=self.TRANQUIL_SPIRIT, effect=self.TRANQUIL_SPIRIT_EFFECT, updater=updater},
  }
  self.removeFuncMap["waterTotem"] = function () DestroyTotem(3) end;
  table.insert(self.setButtons, "waterTotemSet");

  updater = timerUpdaterLib.totemTimerUpdaterFactory:Create(1);

  local flametongueEffect = self.FLAMETONGUE_EFFECT;
  self.actionBarConfList["fireTotem"] = {
    {name=self.FLAMETONGUE, raidEffect=true, effect=function () 
      if environment:GetTalentRankByIndex(16, 1) == 1 then
       return self.TOTEMIC_WRATH_EFFECT;
      end    
      return self.FLAMETONGUE_EFFECT; 
    end, updater=updater},
    {name=self.MAGMA, duration = 20, tickFrequency = 2, raidEffect=true, effect=function () 
      if environment:GetTalentRankByIndex(16, 1) == 1 then
       return self.TOTEMIC_WRATH_EFFECT;
      end    
      return nil; 
    end, updater=updater},
    {name=self.SEARING, duration = function (rank)
          --  if rank >= 7 then
              return 60;
          --  end
          --  return 5*(5+rank);
          end, raidEffect=true, effect=function () 
      if environment:GetTalentRankByIndex(16, 1) == 1 then
       return self.TOTEMIC_WRATH_EFFECT;
      end    
      return nil; 
    end, updater=updater
    },
    {name=self.FIRE_ELEMENTAL, duration = 120, raidEffect=true, effect=function () 
      if environment:GetTalentRankByIndex(16, 1) == 1 then
       return self.TOTEMIC_WRATH_EFFECT;
      end    
      return nil; 
    end,  updater=updater},
  }
  self.removeFuncMap["fireTotem"] = function () DestroyTotem(1) end;
  table.insert(self.setButtons, "fireTotemSet");

  updater = timerUpdaterLib.totemTimerUpdaterFactory:Create(2);

  self.actionBarConfList["earthTotem"] = {
    {name=self.EARTHBIND, duration = 45, tickFrequency = 3, tickStart = .2, updater=updater},
    {name=self.STONECLAW, duration = 15, tickFrequency = 2, life = function (rank)
            local level = UnitLevel("player");
            local base = 0;
            if rank==1 then base = 26;
            elseif rank==2 then base = 96;
            elseif rank==3 then base = 136;
            elseif rank==4 then base = 166;
            elseif rank==5 then base = 246;
            elseif rank==6 then base = 306;
            elseif rank==7 then base = 1114;
            end
            local talentRank = environment:GetTalentRank(self.EARTH_TOTEM_TALENT, 1);
            return (base+3*level)*(4+talentRank)/4;
          end, updater=updater
    },
    {name=self.STONESKIN, raidEffect=true, effect=self.STONESKIN_EFFECT, updater=updater},
    {name=self.STRENGTH_OF_EARTH, raidEffect=true, effect=self.STRENGTH_OF_EARTH_EFFECT, updater=updater},
    {name=self.TREMOR, tickFrequency = 3, updater=updater},
    {name=self.EARTH_ELEMENTAL, duration = 120, updater=updater},
  }
  self.removeFuncMap["earthTotem"] = function () DestroyTotem(2) end;
  table.insert(self.setButtons, "earthTotemSet");
  
  self.actionBarConfList["totemCall"] = {
    {name=self.FIRST_CALL, duration = -1},
    {name=self.SECOND_CALL, duration = -1},
    {name=self.THIRD_CALL, duration = -1},
  }
  self.removeFuncMap["totemCall"] = function ()  end;
  
  local weaponDuration = 1800;

  updater = timerUpdaterLib.weaponTimerUpdaterFactory:Create(true);
  self.actionBarConfList["mainHandWeapon"] = {
    {name=self.WINDFURY_WEAPON, target=Action.MAIN_HAND_WEAPON, duration = weaponDuration, updater=updater},
    {name=self.ROCKBITER_WEAPON, target=Action.MAIN_HAND_WEAPON, duration = weaponDuration, updater=updater},
    {name=self.FLAMETONGUE_WEAPON, target=Action.MAIN_HAND_WEAPON, duration = weaponDuration, updater=updater},
    {name=self.FROSTBRAND_WEAPON, target=Action.MAIN_HAND_WEAPON, duration = weaponDuration, updater=updater},
    {name=self.EARTHLIVING_WEAPON, target=Action.MAIN_HAND_WEAPON, duration = weaponDuration, updater=updater},
  }
  self.removeFuncMap["mainHandWeapon"] = function () CancelItemTempEnchantment(1) end;
  table.insert(self.setButtons, "mainHandWeaponSet");

  updater = timerUpdaterLib.weaponTimerUpdaterFactory:Create(false);
  self.actionBarConfList["offHandWeapon"] = {
    {name=self.WINDFURY_WEAPON, target=Action.OFF_HAND_WEAPON, duration = weaponDuration, updater=updater},
    {name=self.ROCKBITER_WEAPON, target=Action.OFF_HAND_WEAPON, duration = weaponDuration, updater=updater},
    {name=self.FLAMETONGUE_WEAPON, target=Action.OFF_HAND_WEAPON, duration = weaponDuration, updater=updater},
    {name=self.FROSTBRAND_WEAPON, target=Action.OFF_HAND_WEAPON, duration = weaponDuration, updater=updater},
    {name=self.EARTHLIVING_WEAPON, target=Action.OFF_HAND_WEAPON, duration = weaponDuration, updater=updater},
  }
  self.removeFuncMap["offHandWeapon"] = function () CancelItemTempEnchantment(2) end;
  table.insert(self.setButtons, "offHandWeaponSet");

  self.actionBarConfList["shield"] = {
    {name=self.LIGHTNING_SHIELD, duration = 600},
    {name=self.WATER_SHIELD, duration = 600},
    {name=self.EARTH_SHIELD, target="player", duration = 600},
  }
  for _,v in pairs(self.actionBarConfList["shield"]) do
    v.updater = timerUpdaterLib.buffTimerUpdaterFactory:Create(v.name);
  end
  self.removeFuncMap["shield"] = function () 
    CancelUnitBuff("player", self.EARTH_SHIELD);
    CancelUnitBuff("player", self.LIGHTNING_SHIELD);
    CancelUnitBuff("player", self.WATER_SHIELD);
  end;
  table.insert(self.setButtons, "shieldSet");
  
  updater = timerUpdaterLib.friendUniqueBuffTimerUpdaterFactory:Create(self.EARTH_SHIELD, actionLib.spellActionFactory:Create(self.EARTH_SHIELD, "player"));

  self.actionBarConfList["earthShield"] = {
    {name=self.EARTH_SHIELD, duration = 600, updater=updater},
    {name=self.EARTH_SHIELD, target="focus", duration = 600, updater=updater},
    {name=self.EARTH_SHIELD, target="targettarget", duration = 600, updater=updater},
    {name=self.EARTH_SHIELD, target="mouseover", duration = 600, updater=updater},
  }
  table.insert(self.setButtons, "earthShieldSet");

  self.actionBarConfList["ankh"] = {
    {name=self.REINCARNATION, updater=ankhTimerUpdater},
  }

  self.actionBarConfList["natureSwiftness"] = {
    {name=self.NATURE_SWIFTNESS},
  }


  --[[ 
  -----------------
    GUI CREATION PART
  -----------------
  --]]
    
  self.frames = {};

  local messages = {
    started=self.TOTEM_STARTED,
    stopping=self.TOTEM_STOPPING,
    stopped=self.TOTEM_STOPPED,
    countAtZero=self.TOTEM_COUNT_AT_ZERO
  };
  
  local tooltipLabels = {
    actionBarSelectedButton=self.ACTION_BAR_SELECTED_BUTTON,
    addFavorite=self.ADD_FAVORITE,
    removeFavorite=self.REMOVE_FAVORITE, 
    openActionBar=self.OPEN_ACTION_BAR, 
    performAction=self.PERFORM_ACTION, 
    cancelActionResult=self.CANCEL_ACTION_RESULT, 
    actionBarFavoriteButton=self.ACTION_BAR_FAVORITE_BUTTON, 
    exchangeWithNext=self.EXCHANGE_WITH_NEXT, 
    exchangeWithPrevious=self.EXCHANGE_WITH_PREVIOUS, 
    exchangeWithSelected=self.EXCHANGE_WITH_SELECTED, 
    actionBarButton=self.ACTION_BAR_BUTTON, 
    closeActionBarSelectAction=self.CLOSE_ACTION_BAR_SELECT_ACTION, 
    closeActionBarSelectActionPerformAction=self.CLOSE_ACTION_BAR_SELECT_ACTION_PERFORM_ACTION
  };
  
  self.actionBarSkin = {
    
    function () return squareActionBarRendererLib.squareActionBarRendererFactory:Create(tooltipLabels, 30, 30); end,
    
    function () return circleActionBarRendererLib.circleActionBarRendererFactory:Create(tooltipLabels, 30, "Interface\\AddOns\\TotemManager\\textures\\SerenityBase", "Interface\\AddOns\\TotemManager\\textures\\SerenityBaseGreen"); end
    
  }
  
  local rendererCreatorFunc = self.actionBarSkin[self:GetActionBarSkin()];
  
  self.frames["totemActionFrame"] = managerFrameLib.managerFrameFactory:Create(
    self.wowFrame["totemActionFrame"], 
    {"airTotem", "waterTotem", "fireTotem", "earthTotem"},
    self.actionBarConfList,
    "totem",
    self.removeFuncMap,
    totemManagerData["totemActionFrame"],
    messages,
    rendererCreatorFunc,
    280,
    590
  );
  totemManagerData["totemActionFrame"] = self.frames["totemActionFrame"]:GetBackup();
  
  -- global variable totemManagerSet
  local totemManagerSet = CreateFrame("Button", "totemManagerSet", nil, "SecureUnitButtonTemplate");
  totemManagerSet:SetAttribute("type", "macro");
  SecureHandlerWrapScript(totemManagerSet, "OnClick", CreateFrame("Button", nil, nil, "SecureHandlerClickTemplate"), [[
    self:SetAttribute("macrotext", "/click ]]..self.frames["totemActionFrame"]:GetFirstSetButton():GetName()..[[ "..button);
    return true;  
  ]]);


  messages = {
    started=self.BUFF_STARTED,
    stopping=self.BUFF_STOPPING,
    stopped=self.BUFF_STOPPED,
  }; 
    
  self.frames["buffActionFrame"] = managerFrameLib.managerFrameFactory:Create(
    self.wowFrame["buffActionFrame"],
    {"mainHandWeapon", "offHandWeapon", "shield", "earthShield"},
    self.actionBarConfList,
    "buff",
    self.removeFuncMap,
    totemManagerData["buffActionFrame"],
    messages,
    rendererCreatorFunc,
    280,
    390
  );
  totemManagerData["buffActionFrame"] = self.frames["buffActionFrame"]:GetBackup();
  
  -- after totem, buff
  self.frames["totemActionFrame"]:SetNextSetButton(self.frames["buffActionFrame"]:GetFirstSetButton());

  self.frames["longCooldownActionFrame"] = managerFrameLib.managerFrameFactory:Create(
    self.wowFrame["longCooldownActionFrame"],
    {"ankh", "natureSwiftness"},
    self.actionBarConfList,
    "cooldown",
    self.removeFuncMap,
    totemManagerData["longCooldownActionFrame"],
    messages,
    rendererCreatorFunc,
    845,
    615
  );
  totemManagerData["longCooldownActionFrame"] = self.frames["longCooldownActionFrame"]:GetBackup();
  
  
  self.frames["totemCallFrame"] = managerFrameLib.managerFrameFactory:Create(
    self.wowFrame["totemCallFrame"],
    {"totemCall"},
    self.actionBarConfList,
    "buff",
    self.removeFuncMap,
    totemManagerData["totemCallFrame"],
    messages,
    rendererCreatorFunc,
    400,
    625
  );
  totemManagerData["totemCallFrame"] = self.frames["totemCallFrame"]:GetBackup();  
  
  local scale = self:GetScale();
  for _, frame in pairs(self.wowFrame) do 
    frame:SetScale(scale);
    frame:Raise();
  end

  local tcFrame = CreateFrame("Button", "totemicCall", nil, "SecureUnitButtonTemplate");
  tcFrame:SetAttribute("type", "spell");
  tcFrame:SetAttribute("spell", self.TOTEMIC_CALL);
  
  self:UPDATE_BINDINGS();
  
  self:CreateConsole();
end

function TotemManager:PrepareFrame()
  -- frame are created before PLAYER_ENTERING_WORLD.
  self.wowFrame = {
    totemActionFrame = CreateFrame("Frame", "totemActionFrame", parent),
    buffActionFrame = CreateFrame("Frame", "buffActionFrame", parent),
    longCooldownActionFrame = CreateFrame("Frame", "longCooldownActionFrame", parent),
    totemCallFrame = CreateFrame("Frame", "totemCallFrame", parent),
  };  
  for _, frame in pairs(self.wowFrame) do 
    frame:SetMovable(true);
    frame:SetClampedToScreen(true);
  end
end

function TotemManager:initNames()
  for variableName, spellGID in pairs(self.SPELLS_GID) do
    TotemManager[variableName] = GetSpellInfo(spellGID);  
  end
  for variableName, itemGID in pairs(self.ITEMS_GID) do
    TotemManager[variableName] = itemGID; -- since GetItemInfo need local cache filled we use the GID
  end  
end

TotemManager.SPELLS_GID = {
	
  ["TRANQUIL_SPIRIT"] = 87718,
  ["TRANQUIL_SPIRIT_EFFECT"] = 87717,
  ["EARTHBIND"] = 2484,
  ["FIRE_RESISTANCE"] = 8184,
--  ["FROST_RESISTANCE"] = 8181,
  ["FLAMETONGUE"] = 8227,
  ["GROUNDING"] = 8177,
  ["HEALING_STREAM"] = 5394,
  ["MAGMA"] = 8190,
  ["MANA_SPRING"] = 5675,
  ["MANA_TIDE"] = 16190,
--  ["NATURE_RESISTANCE"] = 10595,
--  ["CLEANSING"] = 8170,
  ["SEARING"] = 3599,
 -- ["SENTRY"] = 6495,
  ["STONECLAW"] = 5730,
  ["STONESKIN"] = 8071,
  ["STRENGTH_OF_EARTH"] = 8075,
  ["TREMOR"] = 8143,
  ["WINDFURY"] = 8512,
--  ["WRATH"] = 30706,
  ["WRATH_AIR"] = 3738,
  ["EARTH_ELEMENTAL"] = 2062,
  ["FIRE_ELEMENTAL"] = 2894,

  ["TOTEMIC_CALL"] = 36936,
  ["NATURE_SWIFTNESS"] = 16188,
  ["REINCARNATION"] = 20608,
  ["LIGHTNING_SHIELD"] = 324,
  ["WATER_SHIELD"] = 52127,
  ["EARTH_SHIELD"] = 974,
  ["WINDFURY_WEAPON"] = 8232,
  ["ROCKBITER_WEAPON"] = 8017,
  ["FLAMETONGUE_WEAPON"] = 8024,
  ["FROSTBRAND_WEAPON"] = 8033,
  ["EARTHLIVING_WEAPON"] = 51730,
  
  ["FIRE_RESISTANCE_EFFECT"] = 8185,
--  ["FROST_RESISTANCE_EFFECT"] = 8182,
  ["FLAMETONGUE_EFFECT"] = 52109,
  ["GROUNDING_EFFECT"] = 8178,
  ["ANCIENT_MANA_SPRING_EFFECT"] = 24853,
--  ["NATURE_RESISTANCE_EFFECT"] = 10596,
--  ["SENTRY_EFFECT"] = 6495, -- not a buff ?
  ["STONESKIN_EFFECT"] = 8072,
  ["STRENGTH_OF_EARTH_EFFECT"] = 8076,
  ["WINDFURY_EFFECT"] = 8515,
--  ["WRATH_EFFECT"] = 30708,
  ["TOTEMIC_WRATH_EFFECT"] = 77746,
  ["WRATH_AIR_EFFECT"] = 3738, -- not a buff ?
    
  ["FIRST_CALL"] =  66842,
  ["SECOND_CALL"] = 66843,
  ["THIRD_CALL"] =  66844,

};

TotemManager.ITEMS_GID = {
  ["ENAMORED_WATER_SPIRIT"] = 20503,
  ["ANKH"] = 17030,
};

function TotemManager:PLAYER_ENTERING_WORLD()
  self:initNames();
  if totemManagerData.alreadyActivate4_0_0 then
    -- activate
    self:Activate();
  else
    totemManagerData = {};
--    totemManagerData = self.defaultTotemManagerData;
    self:Activate();  
    totemManagerData.alreadyActivate4_0_0 = true;
    self.console:Print("Successful installation");
  end
  environmentLib.environment:UnregisterEvent("PLAYER_ENTERING_WORLD", self);
end

function TotemManager:VARIABLES_LOADED()
  local _, unitName = UnitClass("player");
  if unitName=="SHAMAN" then
    self:PrepareFrame();
    environmentLib.environment:RegisterEvent("PLAYER_ENTERING_WORLD", self);
  end
  environmentLib.environment:UnregisterEvent("VARIABLES_LOADED", self);
end

-- run the mod (wait for VARIABLES_LOADED event)
environmentLib.environment:RegisterEvent("VARIABLES_LOADED", totemManager);
environmentLib.environment:RegisterEvent("UPDATE_BINDINGS", totemManager);
