--[[
    ActionBarLib
    Copyright (C) 2008 Gael Lalire
    Contact: gael.lalire@laposte.net

    ActionBarLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    ActionBarLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with ActionBarLib.  If not, see <http://www.gnu.org/licenses/>.
--]]

--[[
  The ActionBarRenderer class.
--]]

local actionBarLib = LibStub:NewLibrary("ActionBarLib-1.0", 1);
if not actionBarLib then
  return;
end

-- begin import
local factoryLib = LibStub("FactoryLib-1.0");
local Factory = factoryLib.Factory;
local actionLib = LibStub("ActionLib-1.0");
local Action = actionLib.Action;
local environmentLib = LibStub("EnvironmentLib-1.0");
local environment = environmentLib.environment;
local localizationLib = LibStub("LocalizationLib-1.0");
local localization = localizationLib.localization;
-- end import

local ActionBarRenderer, ActionBarRendererFactory = Factory.CreateAbstractType();
actionBarLib.ActionBarRenderer = ActionBarRenderer;
actionBarLib.ActionBarRendererFactory = ActionBarRendererFactory;

ActionBarRenderer.TOOLTIP_HIDE_IN_COMBAT = 1;
ActionBarRenderer.TOOLTIP_SHOW = 2;
ActionBarRenderer.TOOLTIP_HIDE = 3;

ActionBarRenderer.tooltipVisibility = 1;
ActionBarRenderer.tooltipHelpAdded = true;

function ActionBarRenderer:ApplySkin(secureUnitButton)
  error("ActionBarRenderer:ApplySkin must be overrided");
end

function ActionBarRenderer:UnApplySkin(secureUnitButton)
  error("ActionBarRenderer:UnApplySkin must be overrided");
end

function ActionBarRenderer:ActionChanged(secureUnitButton)
  error("ActionBarRenderer:ActionChanged must be overrided");
end

function ActionBarRenderer:IsTooltipAsked()
  if self.tooltipVisibility == 2 then
    return true;
  elseif self.tooltipVisibility == 3 then
    return false;
  else
    return not InCombatLockdown();
  end
end

function ActionBarRenderer:AddSelectedTooltipHelp(tooltip)
  tooltip:AddLine(localization:Format(self.tooltipLabels.actionBarSelectedButton), 0, 1, 0);
  tooltip:AddLine(localization:Format(self.tooltipLabels.addFavorite, "Alt-leftClick"), 1, 1, 1);
  tooltip:AddLine(localization:Format(self.tooltipLabels.removeFavorite, "Shift-alt-leftClick"), 1, 1, 1);
  tooltip:AddLine(localization:Format(self.tooltipLabels.openActionBar, "RightClick"), 1, 1, 1);
  tooltip:AddLine(localization:Format(self.tooltipLabels.performAction, "LeftClick"), 1, 1, 1);
 -- tooltip:AddLine("Shift-leftClick : Perform action first rank", 1, 1, 1);
  tooltip:AddLine(localization:Format(self.tooltipLabels.cancelActionResult, "Ctrl-leftClick"), 1, 1, 1);
end

function ActionBarRenderer:AddFavoriteTooltipHelp(tooltip)
  tooltip:AddLine(localization:Format(self.tooltipLabels.actionBarFavoriteButton), 0, 1, 0);
  tooltip:AddLine(localization:Format(self.tooltipLabels.exchangeWithNext, "Ctrl-leftclick"), 1, 1, 1);
  tooltip:AddLine(localization:Format(self.tooltipLabels.exchangeWithPrevious, "Shift-ctrl-leftclick"), 1, 1, 1);
  tooltip:AddLine(localization:Format(self.tooltipLabels.exchangeWithSelected, "Rightclick"), 1, 1, 1);
  tooltip:AddLine(localization:Format(self.tooltipLabels.performAction, "LeftClick"), 1, 1, 1);
 -- tooltip:AddLine("Shift-leftclick : Perform action first rank", 1, 1, 1);
end

function ActionBarRenderer:AddOpenTooltipHelp(tooltip)
  tooltip:AddLine(localization:Format(self.tooltipLabels.actionBarButton), 0, 1, 0);
  tooltip:AddLine(localization:Format(self.tooltipLabels.exchangeWithNext, "Ctrl-leftclick"), 1, 1, 1);
  tooltip:AddLine(localization:Format(self.tooltipLabels.exchangeWithPrevious, "Shift-ctrl-leftclick"), 1, 1, 1);
  tooltip:AddLine(localization:Format(self.tooltipLabels.closeActionBarSelectAction, "Rightclick"), 1, 1, 1);
  tooltip:AddLine(localization:Format(self.tooltipLabels.closeActionBarSelectActionPerformAction, "Leftclick"), 1, 1, 1);
 -- tooltip:AddLine("Shift-leftclick : Close the bar, select action and perform action first rank", 1, 1, 1);
end

function ActionBarRendererFactory:Create(tooltipLabels)
  local actionBarRenderer = self:NewInstance();
  actionBarRenderer.tooltipLabels = tooltipLabels;
  return actionBarRenderer;
end


--[[
  The ActionBarState class.
--]]

local ActionBarState, ActionBarStateFactory, actionBarStateFactory = Factory.CreateType();

function ActionBarState:IsOpened()
  return self.selectedVisibility == 0;
end

-- 0,1 open ; 2,3 close ; 1,3 stuck
function ActionBarState:AddFavorite()
  local actionListSize = #self.visibilityList;
  if self.favoriteCount<actionListSize then
    self.favoriteCount = self.favoriteCount + 1;
    local oldVisibility = self.visibilityList[self.favoriteCount];
    if oldVisibility==0 then
      self.visibilityList[self.favoriteCount] = 1;
    elseif oldVisibility==2 then
      self.visibilityList[self.favoriteCount] = 3;      
    end
  end
end

function ActionBarState:RemoveFavorite()
  if self.favoriteCount>0 then
    local oldVisibility = self.visibilityList[self.favoriteCount];
    if oldVisibility==1 then
      self.visibilityList[self.favoriteCount] = 0;
    elseif oldVisibility==3 then
      self.visibilityList[self.favoriteCount] = 2;      
    end
    self.favoriteCount = self.favoriteCount - 1;
  end
end

function ActionBarState:Open()
  if self.selectedVisibility == 0 then
    error("ActionBarState:Open already open");
  end
  self.selectedVisibility = 0;
  self.selectedAction = 0;
  self.selectedPosition = 0;
  local actionListSize = #self.visibilityList;
  for i=1, actionListSize do
    if self.visibilityList[i] == 4 then
      break;
    elseif self.visibilityList[i] == 2 then
      self.visibilityList[i] = 0;
    elseif self.visibilityList[i] == 3 then
      self.visibilityList[i] = 1;
    end
  end
  self.originalSelectedAction = 0;
end

function ActionBarState:Swap(position)
  if self.selectedVisibility == 0 then
    error("ActionBarState:Swap cannot swap while open");
  end  
  self.selectedAction = self.actionList[position];
  self.originalSelectedAction = selectedAction;
  self.actionList[position] = self.actionList[self.selectedPosition];
  self.actionList[self.selectedPosition] = self.selectedAction;
end

function ActionBarState:Close(position)
  if self.selectedVisibility == 1 then
    error("ActionBarState:Close already closed");
  end
  self.selectedPosition = position;
  local actionId = self.actionList[position];
  self.selectedVisibility = 1;
  self.selectedAction = actionId;
  local actionListSize = #self.visibilityList;
  for i=1, actionListSize do
    if self.visibilityList[i] == 4 then
      break;
    elseif self.visibilityList[i] == 0 then
      self.visibilityList[i] = 2;    
    elseif self.visibilityList[i] == 1 then
      self.visibilityList[i] = 3;
    end
  end
  self.originalSelectedAction = actionId;
end

-- do not call directly
function ActionBarState:InsertAction(actionId, position, stucked)
  local actionListSize = #self.actionList;
  local positionOldVisibility = self.visibilityList[position];

  if positionOldVisibility==4 then
    -- insert at the end (imply no need to push button at the end)
    self.actionList[position] = actionId;
    if stucked then
      -- if previous is stuck
      if position == 1 or self.visibilityList[position - 1]%2 == 1 then
        self.favoriteCount = self.favoriteCount + 1;
        if self.selectedVisibility==0 then
          self.visibilityList[position] = 1;
        else
          self.visibilityList[position] = 3;        
        end
      else
        -- we can't be stuck so we becomes like previous
        self.visibilityList[position] = self.visibilityList[position - 1];      
      end
    else
      -- no issue we set our value
      if self.selectedVisibility==0 then
        self.visibilityList[position] = 0;
      else
        self.visibilityList[position] = 2;        
      end
    end
    return;
  end  
  
  for i=1, actionListSize - position do
    self.actionList[actionListSize-i+1] = self.actionList[actionListSize-i];
  end
  self.actionList[position] = actionId;
  local stateReplaced = false;
  
  local positionNewVisibility = positionOldVisibility;

  if stucked then    
    if position == 1 or self.visibilityList[position - 1]%2 == 1 then
      -- if the previous is stucked then we can set our value
      self.favoriteCount = self.favoriteCount + 1;
      if positionOldVisibility == 0 then
        self.visibilityList[position] = 1;
        positionNewVisibility = 1;
      elseif positionOldVisibility == 2 then
        self.visibilityList[position] = 3; 
        positionNewVisibility = 3;     
      end
    end
  end  
  
  local j = position + 1;
  -- first different
  while self.visibilityList[j] == positionNewVisibility do
    j = j + 1;
  end
  local secondPositionVisibility = self.visibilityList[j];

  -- if we get a non empty button
  if secondPositionVisibility~=4 then
    -- look for '4'
    local k = j + 1;
    while self.visibilityList[k] == secondPositionVisibility do
      k = k + 1;
    end
    self.visibilityList[k] = secondPositionVisibility;
    -- if we did not change our state
    if positionNewVisibility == positionOldVisibility then
      self.visibilityList[j] = positionOldVisibility;
    end
  else
    -- we find 4
    self.visibilityList[j] = positionOldVisibility;
  end
end

function ActionBarState:DeleteAction(position)
  local actionListSize = #self.actionList;
  if self.visibilityList[position]==1 or self.visibilityList[position]==3 then
    self.favoriteCount = self.favoriteCount - 1;
  end
  for i=position, actionListSize-1 do
    self.actionList[i] = self.actionList[i+1];
    self.visibilityList[i] = self.visibilityList[i+1];
  end
  self.actionList[actionListSize] = 0;
  self.visibilityList[actionListSize] = 4;  
end

function ActionBarState:EnableAction(actionId)
  local unavailable = self.unavailable[actionId];
  if unavailable==nil then
    error("ActionBarState:EnableAction already enabled");    
  end
  self.totalCount = self.totalCount + 1;
  local pos = unavailable[1];
  local i = pos;
  for _, v in pairs(self.unavailable) do
    if v[1] < unavailable[1] then
      i = i - 1;
    end
  end
  if self.selectedPosition >= i then
    self.selectedPosition = self.selectedPosition + 1;
  end
  self:InsertAction(actionId, i, unavailable[2]);
  self.unavailable[actionId] = nil;
  if self.originalSelectedAction == actionId then
    self:Close(i);
  end
end

function ActionBarState:IsActionEnabled(actionId)
  return (self.unavailable[actionId] == nil);
end

function ActionBarState:DisableAction(actionId)
  local i = 1;
  local actionListSize = #self.actionList;
  while i<=actionListSize and self.actionList[i] ~= actionId do
    i = i + 1;
  end
  if i>actionListSize then
    error("ActionBarState:DisableAction already disabled or inexistant");  
  end
  self.totalCount = self.totalCount - 1;
  local defined = {};
  for _, v in pairs(self.unavailable) do
    defined[v[1]] = true;
  end

  local pos = 0;
  local j = 0; -- j = # button found
  while j<i do
    pos = pos + 1;
    if not defined[pos] then
      j = j + 1; -- there is a real button
    end
  end
  
  local stucked = (self.visibilityList[i]==1 or self.visibilityList[i]==3);
  
  self.unavailable[actionId] = {pos, stucked};
  if self.selectedPosition > i then
    -- if equals then menu will be open so selectedPosition will be lost
    -- that why no decrease with equals (save one operation)
    self.selectedPosition = self.selectedPosition - 1;
  end

  self:DeleteAction(i);
  
  if self.selectedAction == actionId then
    self:Open();
    self.originalSelectedAction = actionId;
  end
end

local function cloneTable(table)
  local result = {};
  for k, v in pairs(table) do
    if type(v) == "table" then
      result[k] = cloneTable(v);
    else
      result[k] = v;
    end
  end
  return result;
end

local function copyTableFrom(table, otherTable)
  for k, v in pairs(otherTable) do
    if type(v) == "table" then
      -- clone because subtable are private and must not be referenced
      table[k] = cloneTable(v);
    else
      table[k] = v;
    end
  end
end

function ActionBarState:Clone()
  local actionBarState = cloneTable(self);
  actionBarStateFactory:Cast(actionBarState);
  return actionBarState;
end

function ActionBarState:CopyFrom(otherState)
  copyTableFrom(self, otherState);
end

function ActionBarStateFactory:Create(actionListSize)
  local actionBarState = self:NewInstance();
  actionBarState.selectedVisibility = 0;
  actionBarState.selectedAction = 0;
  actionBarState.selectedPosition = 0;
  actionBarState.originalSelectedAction = 0;
  actionBarState.favoriteCount = 0;
  -- all are enabled
  actionBarState.totalCount = actionListSize;
  actionBarState.visibilityList = {};
  actionBarState.actionList = {};
  actionBarState.unavailable = {};

  for i=1,actionListSize do
    actionBarState.visibilityList[i] = 0;
    actionBarState.actionList[i] = i;
  end
  
  return actionBarState;
end

--[[
  The ActionBar class.
--]]


local ActionBar, ActionBarFactory, actionBarFactory = Factory.CreateType();
actionBarLib.ActionBar = ActionBar;
actionBarLib.ActionBarFactory = ActionBarFactory;
actionBarLib.actionBarFactory = actionBarFactory;

function ActionBar:GetSetButton()
  return self.setSecureUnitButton;
end

function ActionBar:SetNextSetButton(nextSetButton)
  -- TODO delay creation to out of combat
  if InCombatLockdown() then
    error("Unable to link set button while in combat");
  end
  SecureHandlerSetFrameRef(self.setSecureUnitButton, "nextSetButton", nextSetButton);
end

function ActionBar:ACTIVE_TALENT_GROUP_CHANGED()
  self.specStates[self.backup.spec] = self.state:Clone();
  local spec = tostring(GetActiveTalentGroup());
  local previousState = self.specStates[spec];
  if previousState ~= nil then
    self.state:CopyFrom(previousState);
    self:SpreadState();
    self.specStates[spec] = nil;
  end
  self.backup.spec = spec;
end

local function stringToSnipset(s)
  local snipset;
  if s == nil then
    return "nil";
  end
  return [["]]..s..[["]];
end

-- returns type, spell, item, unit, targetSlot
local function getActionInfo(action)
  if action == nil then
    return nil;
  end
  local rtype = action:GetType();
  local spell, item = nil, nil;
  if rtype == "spell" then
    spell = action:GetValue();
  elseif rtype == "item" then
    item = action:GetValue();
  end
  local unit, targetSlot = nil, nil;
  local target = action:GetTarget();
  if target == Action.MAIN_HAND_WEAPON then
    targetSlot = "16";
  elseif target == Action.OFF_HAND_WEAPON then
    targetSlot = "17";
  elseif target == "target" then
    -- both nil
  else
    unit = target;
  end
  return rtype, spell, item, unit, targetSlot;
end


function ActionBar:SpreadSetState()

  for name, actionBarState in pairs(self.savedSets) do
  
    local rtype, spell, item, unit, targetSlot = getActionInfo(self.actionList[actionBarState.selectedAction]);
 
    local snipset = [[
      local savedSet = newtable();
      local currentSaveTable = newtable();
      currentSaveTable["type"] = ]]..stringToSnipset(rtype)..[[;
      currentSaveTable["spell"] = ]]..stringToSnipset(spell)..[[;
      currentSaveTable["item"] = ]]..stringToSnipset(item)..[[;
      currentSaveTable["unit"] = ]]..stringToSnipset(unit)..[[;
      currentSaveTable["target-slot"] = ]]..stringToSnipset(targetSlot)..[[;
      savedSet.selectedButton = currentSaveTable;
      savedSet.buttons = newtable();
    ]];
    
    for i, action in ipairs(actionBarState.actionList) do
      local rtype, spell, item, unit, targetSlot = getActionInfo(self.actionList[action]);
      snipset = snipset..[[
        currentSaveTable = newtable();
        currentSaveTable["type"] = ]]..stringToSnipset(rtype)..[[;
        currentSaveTable["spell"] = ]]..stringToSnipset(spell)..[[;
        currentSaveTable["item"] = ]]..stringToSnipset(item)..[[;
        currentSaveTable["unit"] = ]]..stringToSnipset(unit)..[[;
        currentSaveTable["target-slot"] = ]]..stringToSnipset(targetSlot)..[[;
        savedSet.buttons[]]..i..[[] = currentSaveTable;
      ]];
    end
    
    for i=actionBarState.totalCount+1, #actionBarState.visibilityList do
      snipset = snipset..[[
        currentSaveTable = newtable();
        savedSet.buttons[]]..i..[[] = currentSaveTable;
      ]];
    end
    
    local openString = "true";
    if not actionBarState:IsOpened() then
      openString = "false";
    end
    
    snipset = snipset..[[
      savedSet.open = ]]..openString..[[;
      savedSet.selectedPosition = ]]..actionBarState.selectedPosition..[[;
      savedSet.favoriteCount = ]]..actionBarState.favoriteCount..[[;
      savedSet.totalCount = ]]..actionBarState.totalCount..[[;      
      savedSets[]]..stringToSnipset(name)..[[] = savedSet;
    ]];
  
    self.header:Execute(snipset);
  end

end

function ActionBar:SpreadState()

  local actionBarState = self.state;
  local opened = actionBarState:IsOpened();

  if actionBarState.selectedAction ~= 0 then
    local selectedAction = self.actionList[actionBarState.selectedAction];
    self.selectedSecureUnitButton:SetAttribute("type", selectedAction:GetType());
    self.selectedSecureUnitButton:SetAttribute(selectedAction:GetType(), selectedAction:GetValue());
    local target = selectedAction:GetTarget();
    if target == Action.MAIN_HAND_WEAPON then
      self.selectedSecureUnitButton:SetAttribute("target-slot", "16");
      self.selectedSecureUnitButton:SetAttribute("unit", nil);
    elseif target == Action.OFF_HAND_WEAPON then
      self.selectedSecureUnitButton:SetAttribute("target-slot", "17");
      self.selectedSecureUnitButton:SetAttribute("unit", nil);
    elseif target == "target" then
      self.selectedSecureUnitButton:SetAttribute("target-slot", nil);
      self.selectedSecureUnitButton:SetAttribute("unit", nil);
    else
      self.selectedSecureUnitButton:SetAttribute("target-slot", nil);
      self.selectedSecureUnitButton:SetAttribute("unit", target);
    end
    self.selectedSecureUnitButton.action = selectedAction;
    self.actionBarRenderer:ActionChanged(self.selectedSecureUnitButton);
    self.selectedSecureUnitButton:Show();
  else
    self.selectedSecureUnitButton.action = nil;
    self.actionBarRenderer:ActionChanged(self.selectedSecureUnitButton);
    self.selectedSecureUnitButton:Hide();
  end

  for i=1, actionBarState.totalCount do    
    local currentAction = self.actionList[actionBarState.actionList[i]];
    self.secureUnitButtonList[i]:SetAttribute("type", currentAction:GetType())
    self.secureUnitButtonList[i]:SetAttribute(currentAction:GetType(), currentAction:GetValue())
    local target = currentAction:GetTarget();
    if target == Action.MAIN_HAND_WEAPON then
      self.secureUnitButtonList[i]:SetAttribute("target-slot", "16");
      self.secureUnitButtonList[i]:SetAttribute("unit", nil);
    elseif target == Action.OFF_HAND_WEAPON then
      self.secureUnitButtonList[i]:SetAttribute("target-slot", "17");
      self.secureUnitButtonList[i]:SetAttribute("unit", nil);
    elseif target == "target" then
      self.secureUnitButtonList[i]:SetAttribute("target-slot", nil);
      self.secureUnitButtonList[i]:SetAttribute("unit", nil);
    else
      self.secureUnitButtonList[i]:SetAttribute("target-slot", nil);
      self.secureUnitButtonList[i]:SetAttribute("unit", target);
    end
    self.secureUnitButtonList[i].action = currentAction;
    self.actionBarRenderer:ActionChanged(self.secureUnitButtonList[i]);
    
    if opened then
      self.secureUnitButtonList[i]:Show();
    else
      if i <= self.state.favoriteCount then
        self.secureUnitButtonList[i]:Show();
      else
        self.secureUnitButtonList[i]:Hide();      
      end
    end        
  end
  for i=actionBarState.totalCount+1, #actionBarState.visibilityList do
    self.secureUnitButtonList[i].action = nil;
    self.actionBarRenderer:ActionChanged(self.secureUnitButtonList[i]);
    self.secureUnitButtonList[i]:Hide();
  end

  local openString = "true";
  if not actionBarState:IsOpened() then
    openString = "false";
  end
  self.header:Execute("open = "..openString.."; selectedPosition = "..actionBarState.selectedPosition.."; favoriteCount = "..actionBarState.favoriteCount.."; totalCount = "..actionBarState.totalCount..";");
   
  self:SpreadSetState();
end

-- change ctrl-click behaviour
function ActionBar:SetSelectedCtrlClickFunc(removeFunc)
  self.selectedSecureUnitButton.removeFunc = removeFunc;  
end

function ActionBar:SetActionEnabled(actionPosition, enabled)
  if enabled then
    self:EnableAction(actionPosition);
  else
    self:DisableAction(actionPosition);
  end
end

-- disable an action (if selected and close we open)
function ActionBar:DisableAction(actionPosition)
  -- TODO : check in combat
  local actionListSize = #self.actionList;
  if actionPosition<1 or actionPosition>actionListSize then
    error("intern error : actionPosition out of range");
  end
    
  self.state:DisableAction(actionPosition);
  for _, savedSet in pairs(self.savedSets) do
    savedSet:DisableAction(actionPosition);
  end
  for _, specState in pairs(self.specStates) do
    specState:DisableAction(actionPosition);  
  end
end

-- enable an action
function ActionBar:EnableAction(actionPosition)
  -- TODO : check in combat
  local actionListSize = #self.actionList;
  if actionPosition<1 or actionPosition>actionListSize then
    error("intern error : actionPosition out of range");
  end
  
  self.state:EnableAction(actionPosition);
  for _, savedSet in pairs(self.savedSets) do
    savedSet:EnableAction(actionPosition);
  end
  for _, specState in pairs(self.specStates) do
    specState:EnableAction(actionPosition);  
  end
end

-- returns {selectedAction=7, selectedVisibility=0, actionList={2, 1, 6, 7, 3, 0, 0}, visibilityList={2, 2, 2, 2, 2, 4, 4}}
-- this state can be given in constructor (actionBarState)
function ActionBar:GetBackup()
  return self.backup;
end

function ActionBar:ACTION_AVAILABILITY_CHANGED(actionPosition, available)
  if available then
    self:EnableAction(actionPosition);
  else
    self:DisableAction(actionPosition);
  end
  self:SpreadState();
end

function ActionBar:SetActionBarRenderer(actionBarRenderer)

  -- UnApply old skin
  self.actionBarRenderer:UnApplySkin(self.selectedSecureUnitButton);
  for _, secureUnitButton in ipairs(self.secureUnitButtonList) do
    self.actionBarRenderer:UnApplySkin(secureUnitButton);  
  end

  -- Apply new skin
  self.actionBarRenderer = actionBarRenderer;  
  actionBarRenderer:ApplySkin(self.selectedSecureUnitButton);
  actionBarRenderer:ActionChanged(self.selectedSecureUnitButton);
  for _, secureUnitButton in ipairs(self.secureUnitButtonList) do
    actionBarRenderer:ApplySkin(secureUnitButton);
    actionBarRenderer:ActionChanged(secureUnitButton);
  end

end

--[[
  The ActionBarFactory class.
--]]

function ActionBarFactory:CreateSecureUnitButton(buttonName, parent, actionBar, position)
  local secureUnitButton = CreateFrame("Button", buttonName, parent, "SecureUnitButtonTemplate");
  secureUnitButton.actionBar = actionBar;
  secureUnitButton.borderTexture = secureUnitButton:CreateTexture(nil, "OVERLAY");
  
  secureUnitButton.targetText = secureUnitButton:CreateFontString(nil, "ARTWORK", "NumberFontNormal");
  secureUnitButton.targetText:SetAllPoints();
  
  secureUnitButton.cooldown = CreateFrame("Cooldown", nil, secureUnitButton, "CooldownFrameTemplate");  
    
  secureUnitButton.position = position;
  
  secureUnitButton:SetScript("OnEnter", function (button)
    local actionBarRenderer = button.actionBar.actionBarRenderer;
    if actionBarRenderer:IsTooltipAsked() then
      actionBarRenderer:UpdateTooltip(button);      
    end
  end);
  
  if position~=0 then
    secureUnitButton:SetScript("PreClick", function (button, click)
      local i = button.position;
      local actionBar = button.actionBar;
      local state = actionBar.state;
      local currentAction = state.actionList[i];
      local currentButton = button;
      if IsControlKeyDown() then
        local tmpButton = nil;
        local tmpAction;        
        if IsShiftKeyDown() then
          if i<state.totalCount then
            tmpButton = actionBar.secureUnitButtonList[i+1];
            tmpAction = state.actionList[i+1];
            state.actionList[i+1] = currentAction;
          end
        else
          if i>0 then
            tmpButton = actionBar.secureUnitButtonList[i-1];
            tmpAction = state.actionList[i-1];
            state.actionList[i-1] = currentAction;
          end
        end
        if tmpButton~=nil then
          state.actionList[i] = tmpAction;
          tmpButton.action = actionBar.actionList[currentAction];
          actionBar.actionBarRenderer:ActionChanged(tmpButton, actionBar.actionList[currentAction]);
          currentButton.action = actionBar.actionList[tmpAction];
          actionBar.actionBarRenderer:ActionChanged(currentButton);
        end
      elseif IsAltKeyDown() then
        -- nothing to do
      else
        local action = actionBar.actionList[currentAction];
        if state:IsOpened() then
          actionBar.selectedSecureUnitButton.action = action;
          actionBar.actionBarRenderer:ActionChanged(actionBar.selectedSecureUnitButton);      
          state:Close(i);
        else
          if click=="RightButton" then
            state:Swap(i);
            actionBar.selectedSecureUnitButton.action = action;
            actionBar.actionBarRenderer:ActionChanged(actionBar.selectedSecureUnitButton);
            currentButton.action = actionBar.actionList[state.actionList[i]];
            actionBar.actionBarRenderer:ActionChanged(currentButton);
            actionBar.secureUnitButtonList[state.selectedPosition].action = actionBar.actionList[state.actionList[state.selectedPosition]];
            actionBar.actionBarRenderer:ActionChanged(actionBar.secureUnitButtonList[state.selectedPosition]);            
          end
        end
      end
    end);
  
  else
  
    secureUnitButton:SetScript("PreClick", function (button, click)
      local actionBar = button.actionBar;
      local state = actionBar.state;
      if IsControlKeyDown() then
        -- remove action (no effet on state)
        if click=="LeftButton" and button.removeFunc~=nil then 
          button.removeFunc();
        end
      elseif IsAltKeyDown() then
        -- state add / remove favorite
        if IsShiftKeyDown() then
          state:RemoveFavorite();
        else
          state:AddFavorite();        
        end
      else
        if click=="RightButton" then
          state:Open();
        end
      end
    end);
    
  end

  secureUnitButton:SetScript("OnLeave", function ()
    GameTooltip:Hide();
  end);
  
  return secureUnitButton;
end

-- Name of buttons : [buttonName]Selected [buttonName]Set [buttonName][Number]
function ActionBarFactory:Create(parent, buttonName, actionList, actionBarRenderer, backup)
  local actionBar = self:NewInstance();

  actionBar.actionList = actionList;
  
  local actionListSize = #actionList;
  
  if actionListSize == 0 then
    return nil;
  end
   
  local spec = tostring(GetActiveTalentGroup());
  local actionBarState;
  -- this algo keep only available action but not all available action ()  
  if backup == nil then
    backup = {};
    actionBarState = actionBarStateFactory:Create(actionListSize);
    backup.state = actionBarState;
    backup.savedSets = {};
    backup.specStates = {};
    backup.spec = spec;
  else
    actionBarState = backup.state;
    actionBarStateFactory:Cast(actionBarState);
    for _, savedSet in pairs(backup.savedSets) do
      actionBarStateFactory:Cast(savedSet);
    end
    for _, specState in pairs(backup.specStates) do
      actionBarStateFactory:Cast(specState);
    end
  end
  actionBar.backup = backup;
  actionBar.state = actionBarState;
  actionBar.savedSets = backup.savedSets;
  actionBar.specStates = backup.specStates;
  
  -- latest spec is different than current spec
  if backup.spec ~= spec then
    backup.specStates[backup.spec] = actionBarState:Clone();
    local previousState = backup.specStates[spec];
    if previousState ~= nil then
      actionBarState:CopyFrom(previousState);
      backup.specStates[spec] = nil;
    end    
    backup.spec = spec;
  end
  
  for i=1, actionListSize do
    action = actionList[i];
    local actionAvailable = action:IsAvailable();
    local actionEnabled = actionBarState:IsActionEnabled(i);
    --     if actionEnabled ~= actionAvailable then BUG ??
    if actionEnabled and not actionAvailable then
      actionBar:SetActionEnabled(i, actionAvailable);
    end
    if not actionEnabled and actionAvailable then
       actionBar:SetActionEnabled(i, actionAvailable);
    end
  end
    
  local parent = CreateFrame("Button", nil, parent, "SecureHandlerClickTemplate");    
  actionBar.header = parent;
  
  local clickSnipset = [[
    local currentButton = self;
    if IsControlKeyDown() then
      local i=0;
      while buttons[i]~=self do
        i = i + 1;
      end      
      local tmpButton = nil;
      if IsShiftKeyDown() then
        if i<totalCount then
          tmpButton = buttons[i + 1];
        end
      else
        if i>0 then
          tmpButton = buttons[i - 1];
        end
      end
      if tmpButton~=nil then
        local tmpType = tmpButton:GetAttribute("type");
        local tmpSpell = tmpButton:GetAttribute("spell");
        local tmpItem = tmpButton:GetAttribute("item");
        local tmpUnit = tmpButton:GetAttribute("unit");
        local tmpSlot = tmpButton:GetAttribute("target-slot");
        tmpButton:SetAttribute("type", currentButton:GetAttribute("type"));
        tmpButton:SetAttribute("spell", currentButton:GetAttribute("spell"));
        tmpButton:SetAttribute("item", currentButton:GetAttribute("item"));
        tmpButton:SetAttribute("unit", currentButton:GetAttribute("unit"));
        tmpButton:SetAttribute("target-slot", currentButton:GetAttribute("target-slot"));        
        currentButton:SetAttribute("type", tmpType);
        currentButton:SetAttribute("spell", tmpSpell);
        currentButton:SetAttribute("item", tmpItem);
        currentButton:SetAttribute("unit", tmpUnit);
        currentButton:SetAttribute("target-slot", tmpSlot);
      end
      return false;
    elseif IsAltKeyDown() then
        -- nothing with alt-key
        return false;
    else
      if button=="LeftButton" or button=="RightButton" then
        if open then
          local i=0;
          while buttons[i]~=self do
            i = i + 1;
          end      
          selectedPosition = i;
          for i=favoriteCount+1,totalCount do
            buttons[i]:Hide();
          end
          selectedButton:SetAttribute("type", currentButton:GetAttribute("type"));
          selectedButton:SetAttribute("spell", currentButton:GetAttribute("spell"));
          selectedButton:SetAttribute("item", currentButton:GetAttribute("item"));
          selectedButton:SetAttribute("unit", currentButton:GetAttribute("unit"));
          selectedButton:SetAttribute("target-slot", currentButton:GetAttribute("target-slot"));
          selectedButton:Show();
          open = false;
        else
          if button=="RightButton" then
            local fromButton = self;
            local toButton = selectedButton;
            toButton:SetAttribute("type", fromButton:GetAttribute("type"));
            toButton:SetAttribute("spell", fromButton:GetAttribute("spell"));
            toButton:SetAttribute("item", fromButton:GetAttribute("item"));
            toButton:SetAttribute("unit", fromButton:GetAttribute("unit"));
            toButton:SetAttribute("target-slot", fromButton:GetAttribute("target-slot"));
            fromButton = buttons[selectedPosition];
            toButton = self;
            toButton:SetAttribute("type", fromButton:GetAttribute("type"));
            toButton:SetAttribute("spell", fromButton:GetAttribute("spell"));
            toButton:SetAttribute("item", fromButton:GetAttribute("item"));
            toButton:SetAttribute("unit", fromButton:GetAttribute("unit"));
            toButton:SetAttribute("target-slot", fromButton:GetAttribute("target-slot"));
            fromButton = selectedButton;
            toButton = buttons[selectedPosition];
            toButton:SetAttribute("type", fromButton:GetAttribute("type"));
            toButton:SetAttribute("spell", fromButton:GetAttribute("spell"));
            toButton:SetAttribute("item", fromButton:GetAttribute("item"));
            toButton:SetAttribute("unit", fromButton:GetAttribute("unit"));
            toButton:SetAttribute("target-slot", fromButton:GetAttribute("target-slot"));
          end
        end
        -- do not cast if RightButton
        return button=="LeftButton";      
      else
        return false;      
      end
    end
  ]];

  local selectedClickSnipset = [[
    if IsControlKeyDown() then
      -- try to avoid action with a post click method
      return false;
    elseif IsAltKeyDown() then
      if IsShiftKeyDown() then
        -- remove favorite
        if favoriteCount>0 then
          buttons[favoriteCount]:Hide();
          favoriteCount = favoriteCount - 1;
        end
      else
        -- add favorite
        if favoriteCount<totalCount then
          favoriteCount = favoriteCount + 1;
          buttons[favoriteCount]:Show();
        end
      end
      return false;
    else
      if button=="LeftButton" then
        return true;      
      elseif button=="RightButton" then
        selectedPosition = 0;
        for i=favoriteCount+1,totalCount do
          buttons[i]:Show();
        end
        self:Hide();
        open = true;
        return false;
      else
        return false;      
      end
    end
  ]];

  local setClickSnipset = [[
    local _, _, action, number = string.find(button, "([SDR])(%d+)");
    if action=="S" then
      -- save set
      local savedSet = newtable();
      local button = selectedButton;
      local currentSaveTable = newtable();
      currentSaveTable["type"] = button:GetAttribute("type");
      currentSaveTable["spell"] = button:GetAttribute("spell");
      currentSaveTable["item"] = button:GetAttribute("item");
      currentSaveTable["unit"] = button:GetAttribute("unit");
      currentSaveTable["target-slot"] = button:GetAttribute("target-slot");
      savedSet.selectedButton = currentSaveTable;
      savedSet.buttons = newtable();
      for i, button in ipairs(buttons) do
        currentSaveTable = newtable();
        currentSaveTable["type"] = button:GetAttribute("type");
        currentSaveTable["spell"] = button:GetAttribute("spell");
        currentSaveTable["item"] = button:GetAttribute("item");
        currentSaveTable["unit"] = button:GetAttribute("unit");
        currentSaveTable["target-slot"] = button:GetAttribute("target-slot");
        savedSet.buttons[i] = currentSaveTable;
      end
      savedSet.open = open;
      savedSet.selectedPosition = selectedPosition;
      savedSet.favoriteCount = favoriteCount;
      savedSet.totalCount = totalCount;      
      savedSets[number] = savedSet;
    elseif action=="D" then
      -- delete set
      savedSets[number] = nil;
    elseif action=="R" then
      -- restore set
      local savedSet = savedSets[number];
      if savedSet ~= nil then
        open = savedSet.open;
        selectedPosition = savedSet.selectedPosition;
        favoriteCount = savedSet.favoriteCount;
        totalCount = savedSet.totalCount;
        local button = selectedButton;
        local currentSaveTable = savedSet.selectedButton;
        button:SetAttribute("type", currentSaveTable["type"]);
        button:SetAttribute("spell", currentSaveTable["spell"]);
        button:SetAttribute("item", currentSaveTable["item"]);
        button:SetAttribute("unit", currentSaveTable["unit"]);
        button:SetAttribute("target-slot", currentSaveTable["target-slot"]);
        if open then
          button:Hide();
        else
          button:Show();
        end
        for i, button in ipairs(buttons) do
          currentSaveTable = savedSet.buttons[i];
          button:SetAttribute("type", currentSaveTable["type"]);
          button:SetAttribute("spell", currentSaveTable["spell"]);
          button:SetAttribute("item", currentSaveTable["item"]);
          button:SetAttribute("unit", currentSaveTable["unit"]);
          button:SetAttribute("target-slot", currentSaveTable["target-slot"]);
          if i <= totalCount and (open or i <= favoriteCount) then
            button:Show();
          else
            button:Hide();
          end
        end
      end
    end
    local nextSetButton = self:GetFrameRef("nextSetButton");
    if nextSetButton == nil then
      return false;
    else
      self:SetAttribute("type", "macro");
      self:SetAttribute("macrotext", "/click "..nextSetButton:GetName().." "..button);
      return true;
    end
  ]];

  local initSnipset = [[
    local allChilds = newtable(owner:GetChildren());
    local size = #allChilds - 2;
    setButton = allChilds[1];
    selectedButton = allChilds[2];
    buttons = newtable();
    for i=1, size do
      buttons[i] = allChilds[i+2];    
    end
    savedSets = newtable();
  ]];
  
  local setSecureUnitButton = self:CreateSecureUnitButton(buttonName.."Set", parent, actionBar, 0);
  actionBar.setSecureUnitButton = setSecureUnitButton;
  setSecureUnitButton:Hide();
  setSecureUnitButton:RegisterForClicks("LeftButtonUp", "RightButtonUp");
  SecureHandlerWrapScript(setSecureUnitButton, "OnClick", parent, setClickSnipset);

  setSecureUnitButton.actionBar = actionBar;  
  setSecureUnitButton:SetScript("PreClick", function (button, click)
    local actionBar = button.actionBar;
    local state = actionBar.state;
    local savedSets = actionBar.savedSets;
    local _, _, action, number = string.find(click, "([SDR])(%d+)");
    if action=="S" then
      -- save set
      savedSets[number] = state:Clone();
    elseif action=="D" then
      -- delete set
      savedSets[number] = nil;
    elseif action=="R" then
      -- restore set
      local savedSet = savedSets[number];
      if savedSet ~= nil then
        state:CopyFrom(savedSet);
        actionBar.selectedSecureUnitButton.action = actionBar.actionList[state.selectedAction];
        actionBar.actionBarRenderer:ActionChanged(actionBar.selectedSecureUnitButton);
        for i, secureUnitButton in ipairs(actionBar.secureUnitButtonList) do
         secureUnitButton.action = actionBar.actionList[state.actionList[i]];
         actionBar.actionBarRenderer:ActionChanged(secureUnitButton);
        end
      end
    end
  end);
     
  local selectedSecureUnitButton = self:CreateSecureUnitButton(buttonName.."Selected", parent, actionBar, 0);
  actionBar.selectedSecureUnitButton = selectedSecureUnitButton;
  selectedSecureUnitButton:Hide();
  actionBarRenderer:ApplySkin(selectedSecureUnitButton);  
  selectedSecureUnitButton:RegisterForClicks("LeftButtonUp", "RightButtonUp");
  SecureHandlerWrapScript(selectedSecureUnitButton, "OnClick", parent, selectedClickSnipset);

  -- tmp variable to complete link between buttons
  local secureUnitButtonList = {}; -- visible buttons (created by renderer)
  actionBar.secureUnitButtonList = secureUnitButtonList;
  
  for i=1,actionListSize do
    -- keep actionListSize state (one by action)  
    local secureUnitButton = self:CreateSecureUnitButton(buttonName..i, parent, actionBar, i);
    secureUnitButtonList[i] = secureUnitButton;
    actionBarRenderer:ApplySkin(secureUnitButton);
    secureUnitButton:RegisterForClicks("LeftButtonUp", "RightButtonUp");
    SecureHandlerWrapScript(secureUnitButton, "OnClick", parent, clickSnipset);
  end
      
  actionBar.actionBarRenderer = actionBarRenderer;
  
  parent:Execute(initSnipset);
      
  -- spread state
  actionBar:SpreadState();    
  
  for i=1,actionListSize do
    local action = actionList[i];
    -- can be directly enable, that why we register at the end
    action:RegisterEvent("ACTION_AVAILABILITY_CHANGED", actionBar, i);
  end
  
  environment:RegisterEvent("ACTIVE_TALENT_GROUP_CHANGED", actionBar);
  
  return actionBar;  
end

