--[[
    ActionLib
    Copyright (C) 2008 Gael Lalire
    Contact: gael.lalire@laposte.net

    ActionLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    ActionLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with ActionLib.  If not, see <http://www.gnu.org/licenses/>.
--]]

local actionLib = LibStub:NewLibrary("ActionLib-1.0", 1);
if not actionLib then
  return;
end

-- begin import
local factoryLib = LibStub("FactoryLib-1.0");
local Factory = factoryLib.Factory;
local eventHandlerLib = LibStub("EventHandlerLib-1.0");
local EventHandlerFactory = eventHandlerLib.EventHandlerFactory;
local environmentLib = LibStub("EnvironmentLib-1.0");
local environment = environmentLib.environment;
-- end import


--[[
  The Action class.
--]]

-- Action extends EventHandler
local Action, ActionFactory = Factory.CreateAbstractType(EventHandlerFactory)

actionLib.Action = Action;
actionLib.ActionFactory = ActionFactory;

Action.MAIN_HAND_WEAPON = "mhw";
Action.OFF_HAND_WEAPON = "ohw";

-- type of action
function Action:GetType()
  return self.type;
end

-- shift-type[-button] will be set to value
function Action:GetShiftValue()
  return self.shiftValue;
end

-- type[-button] will be set to value
function Action:GetValue()
  return self.value;
end

function Action:GetTarget()
  return self.target;
end

function Action:GetTexture()
  return self.texture;
end

function Action:GetCooldown()
  error("Action:GetCooldown must be overrided");
end

function Action:ShowInTooltip(tooltip)
  error("Action:ShowInTooltip must be overrided");
end

function Action:CanBeCast()
  error("Action:CanBeCast must be overrided");
end

function Action:IsAvailable()
  self:CheckAvailability(); -- the check is not necessary if event are registred
  return self.available;
end

-- never send an event but may change self.available if availability changed
function Action:CheckAvailability()
  error("Action:CheckAvailability must be overrided");
end

--[[
  The ActionFactory class.
--]]

-- type of action
function ActionFactory:Create(actionType, value, shiftValue, target, texture)
  local action = self:NewInstance();
  action.type = actionType;
  action.value = value;
  action.shiftValue = shiftValue;
  action.target = target;
  action.texture = texture;
  return action;
end

--[[
  The SpellAction class.
--]]

-- SpellAction extends Action
local SpellAction, SpellActionFactory, spellActionFactory = Factory.CreateType(ActionFactory);

actionLib.SpellAction = SpellAction;
actionLib.SpellActionFactory = SpellActionFactory;
actionLib.spellActionFactory = spellActionFactory;

SpellAction.RANK_ONE = "(Rank 1)";
local locale = GetLocale();
if locale == "frFR" or locale == "deDE" then
  SpellAction.RANK_ONE = "(Rang 1)";
elseif locale == "enES" then
  SpellAction.RANK_ONE = "(Rango 1)";
elseif locale == "ruRU" then
  SpellAction.RANK_ONE = "(Уровень 1)";
end

-- send an event if availability changed
function SpellAction:CheckAvailability()
  if self.available then
    if environment:GetSpellMaxRank(self.spellName)==0 then
      -- if we forget the spell
      self.available = false;
    end
  else
    if environment:GetSpellMaxRank(self.spellName)~=0 then
      -- if we learn the spell
      self.available = true;
      -- load data if not loaded
      if self.value == nil then
        local name, rank, icon = GetSpellInfo(self.spellName);
        self.type = "spell";
        self.value = name;
        if rank=="" then
          self.shiftValue = name;    
        else
          self.shiftValue = name..SpellAction.RANK_ONE;
        end
        self.texture = icon;
      end
    end
  end
end

function SpellAction:OnUsed(eventName)
  if eventName == "ACTION_AVAILABILITY_CHANGED" then
    if self.available then
      environment:RegisterEvent("SPELLS_FORGOTTEN", self);
    else
      environment:RegisterEvent("SPELLS_LEARNT", self);
    end
    self.waitForAvailable = true;
  elseif eventName == "ACTION_PERFORMED" then
    environment:RegisterEvent("UNIT_SPELLCAST_SENT", self);
  elseif eventName == "COOLDOWN_UPDATED" then
    environment:RegisterEvent("SPELL_UPDATE_COOLDOWN", self);
  end
end

function SpellAction:OnUnused(eventName)
  if eventName == "ACTION_AVAILABILITY_CHANGED" then
    if self.available then
      environment:UnregisterEvent("SPELLS_FORGOTTEN", self);
    else
      environment:UnregisterEvent("SPELLS_LEARNT", self);    
    end
    self.waitForAvailable = false;
  elseif eventName == "ACTION_PERFORMED" then
    if self.waitForCastResult then
      environment:UnregisterEvent("UNIT_SPELLCAST_STOP", self);      
      environment:UnregisterEvent("UNIT_SPELLCAST_SUCCEEDED", self);
      self.waitForCastResult = false;
    elseif self.waitForWeaponHand then
      environment:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED", self);
      self.waitForWeaponHand = false;
    elseif self.waitForDurationUpdate then
      environment:UnregisterEvent("UPDATE", self);
      self.waitForDurationUpdate = false;
    end
    environment:UnregisterEvent("UNIT_SPELLCAST_SENT", self);      
  elseif eventName == "COOLDOWN_UPDATED" then
    environment:UnregisterEvent("SPELL_UPDATE_COOLDOWN", self);
  end
end

function SpellAction:SPELL_UPDATE_COOLDOWN()
  if self:IsAvailable() then
    self:Fire("COOLDOWN_UPDATED", self);
  end
end


function SpellAction:UNIT_SPELLCAST_SENT(unit, spellName, rank, target)
  if unit ~= "player" or spellName ~= self.spellName then
    return;
  end
  -- check because mana spring can have or not a rank
  if rank=="" and not environment:IsUnrankedSpell(spellName) then
    return;
  end
  local validTarget = false;
  if self.target=="target" and (target=="" or UnitName("target")==nil) then
    validTarget = true;
  elseif self.target == self.MAIN_HAND_WEAPON or self.target == self.OFF_HAND_WEAPON then
    validTarget = true;
  else
    if environment:UnitFullName(self.target)==target then
      validTarget = true;
    end
  end
  if validTarget then
    self.targetName = target;
    environment:RegisterEvent("UNIT_SPELLCAST_STOP", self);
    environment:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED", self);
    self.waitForCastResult = true;
  end  
end

function SpellAction:UNIT_SPELLCAST_STOP(unit, spellName)
  if not self.waitForCastResult then
    return;
  end
  if unit=="player" and spellName==self.spellName then
    environment:UnregisterEvent("UNIT_SPELLCAST_STOP", self);      
    environment:UnregisterEvent("UNIT_SPELLCAST_SUCCEEDED", self);
    self.waitForCastResult = false;
  end
end

function SpellAction:UNIT_SPELLCAST_SUCCEEDED(unit, spellName, rank)
  if not self.waitForCastResult then
    return;
  end
  if unit=="player" and spellName==self.spellName then
    environment:UnregisterEvent("UNIT_SPELLCAST_STOP", self);
    environment:UnregisterEvent("UNIT_SPELLCAST_SUCCEEDED", self);
    self.waitForCastResult = false;
    if self.target == self.MAIN_HAND_WEAPON or self.target == self.OFF_HAND_WEAPON then
      environment:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED", self);
      self.waitForWeaponHand = true;
      self.rankSucceeded = environment:RankOfString(rank);
    else
      self:Fire("ACTION_PERFORMED", self, environment:RankOfString(rank), self.targetName);
    end
  end
end

function SpellAction:COMBAT_LOG_EVENT_UNFILTERED(_, event, _, _, _, _, unit, _, spellName, weapon)
  if not self.waitForWeaponHand then
    return;
  end
  if unit == UnitName("player") and event == "ENCHANT_APPLIED" then
    self.waitForWeaponHand = false;
    local rankSucceeded = self.rankSucceeded;
    self.rankSucceeded = nil;    
    environment:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED", self);

    local _, itemLink = GetItemInfo(weapon);
    local item = environment:GetItemGID(itemLink);
    local mainHandLink = environment:GetWeaponLink(true);
    local offHandLink = environment:GetWeaponLink(false);
    
    if self.target == self.MAIN_HAND_WEAPON then
      if mainHandLink ~= nil then
        if offHandLink == nil then
          self:Fire("ACTION_PERFORMED", self, rankSucceeded, self.targetName);      
        else
          -- here both main hand and offhand enchant
          local mainHand = environment:GetItemGID(mainHandLink);
          local offHand = environment:GetItemGID(offHandLink);
          if mainHand ~= offHand then
            if mainHand == item then
              self:Fire("ACTION_PERFORMED", self, rankSucceeded, self.targetName);      
            end
          else
            -- still unknown target ...
            local hasMainHandEnchant, mainHandExpirationMilli, _, hasOffHandEnchant, offHandExpirationMilli = GetWeaponEnchantInfo();
            local mainHandExpiration, offHandExpiration = (mainHandExpirationMilli or 0)/1000, (offHandExpirationMilli or 0)/1000;
            if not hasMainHandEnchant then
              self:Fire("ACTION_PERFORMED", self, rankSucceeded, self.targetName);
            else
              if hasOffHandEnchant then
                -- both weapons can be target ...
                self.waitForDurationUpdate = true;
                local time = GetTime();
                self.mainHandExpirationTime = time + mainHandExpiration;
                self.offHandExpirationTime = time + offHandExpiration;
                environment:RegisterEvent("UPDATE", self);
              end
            end
          end
        end
      end
    elseif self.target == self.OFF_HAND_WEAPON then
      if offHandLink ~= nil then
        if mainHandLink == nil then
          self:Fire("ACTION_PERFORMED", self, rankSucceeded, self.targetName);      
        else
          -- here both main hand and offhand enchant
          local mainHand = environment:GetItemGID(mainHandLink);
          local offHand = environment:GetItemGID(offHandLink);
          if mainHand ~= offHand then
            if offHand == item then
              self:Fire("ACTION_PERFORMED", self, rankSucceeded, self.targetName);      
            end
          else
            -- still unknown target ...
            local hasMainHandEnchant, mainHandExpirationMilli, _, hasOffHandEnchant, offHandExpirationMilli = GetWeaponEnchantInfo();
            local mainHandExpiration, offHandExpiration = (mainHandExpirationMilli or 0)/1000, (offHandExpirationMilli or 0)/1000;
            if hasMainHandEnchant and not hasOffHandEnchant then
              self:Fire("ACTION_PERFORMED", self, rankSucceeded, self.targetName);
            else
              if hasMainHandEnchant then
                -- both weapons can be target ...
                self.waitForDurationUpdate = true;
                local time = GetTime();
                self.mainHandExpirationTime = time + mainHandExpiration;
                self.offHandExpirationTime = time + offHandExpiration;
                environment:RegisterEvent("UPDATE", self);
              end
            end
          end
        end
      end
    end
  end
end

function SpellAction:UPDATE()
  if not self.waitForDurationUpdate then
    return;
  end
  local _, mainHandExpirationMilli, _, _, offHandExpirationMilli = GetWeaponEnchantInfo();
  local mainHandExpiration, offHandExpiration = (mainHandExpirationMilli or 0)/1000, (offHandExpirationMilli or 0)/1000;
  local time = GetTime();
  
  local diffMainHand = math.abs(self.mainHandExpirationTime - time - mainHandExpiration);
  local diffOffHand = math.abs(self.offHandExpirationTime - time - offHandExpiration);
  
  -- Global cooldown 1.5 then .5 is impossible update
  if diffMainHand > .5 or diffOffHand > .5 then
    -- updated
    self.mainHandExpirationTime = nil;
    self.offHandExpirationTime = nil;
    self.waitForDurationUpdate = false;
    environment:UnregisterEvent("UPDATE", self);
    if diffMainHand > diffOffHand then
      -- main hand updated
      if self.target == self.MAIN_HAND_WEAPON then
        self:Fire("ACTION_PERFORMED", self, rankSucceeded, self.targetName);
      end
    else
      -- off hand updated
      if self.target == self.OFF_HAND_WEAPON then
        self:Fire("ACTION_PERFORMED", self, rankSucceeded, self.targetName);
      end    
    end
  end
end

function SpellAction:SPELLS_LEARNT()
  if not self.waitForAvailable then
    return;
  end
  self:CheckAvailability();
  if self.available then
    environment:UnregisterEvent("SPELLS_LEARNT", self);
    environment:RegisterEvent("SPELLS_FORGOTTEN", self);
    self:Fire("ACTION_AVAILABILITY_CHANGED", true);  
  end  
end

function SpellAction:SPELLS_FORGOTTEN()
  if not self.waitForAvailable then
    return;
  end
  self:CheckAvailability();
  if not self.available then
    environment:UnregisterEvent("SPELLS_FORGOTTEN", self);
    environment:RegisterEvent("SPELLS_LEARNT", self);
    self:Fire("ACTION_AVAILABILITY_CHANGED", false);
  end
end

function SpellAction:GetCooldown()
  return GetSpellCooldown(environment:GetSpellId(self.value), BOOKTYPE_SPELL);
end

function SpellAction:ShowInTooltip(tooltip)
  tooltip:SetSpellBookItem(environment:GetSpellId(self.value), BOOKTYPE_SPELL);
end

function SpellAction:CanBeCast()
  return (IsUsableSpell(self.value));
end

--[[
  The ItemAction class.
--]]

-- ItemAction extends Action
local ItemAction, ItemActionFactory, itemActionFactory = Factory.CreateType(ActionFactory);

actionLib.ItemAction = ItemAction;
actionLib.ItemActionFactory = ItemActionFactory;
actionLib.itemActionFactory = itemActionFactory;

-- send an event if availability changed
function ItemAction:CheckAvailability()
  if IsEquippedItem(self.itemGID) then
    if not self.available then
      -- if first available (ie self.value == nil then fill other data)    
      if self.value == nil then
        local name, itemLink, _, _, _, _, _, _, _, itemTexture = GetItemInfo(self.itemGID);
        if name==nil then
          return; -- we do not changed availability while we can't access to item info
        end
        self.itemLink = itemLink; -- keep item link info for tooltip
        self.type = "item";
        self.value = name;
        self.shiftValue = name;
        self.texture = itemTexture;
      end
      self.available = true;
    end
  else
    if self.available then
      self.available = false;
    end    
  end
end

function ItemAction:UNIT_INVENTORY_CHANGED()
  if not self.waitForAvailable then
    return;
  end
  if self.available then
    self:CheckAvailability();
    if not self.available then
      self:Fire("ACTION_AVAILABILITY_CHANGED", false);
    end
  else
    self:CheckAvailability();
    if self.available then
      self:Fire("ACTION_AVAILABILITY_CHANGED", true);
    end
  end
end


function ItemAction:UNIT_SPELLCAST_SENT(unit, spellName, rank, target)
  if unit ~= "player" then
    return;
  end
  -- a item spell has no rank
  local itemSpellName, itemSpellRank = GetItemSpell(self.itemGID);
  if rank~=itemSpellRank or spellName~=itemSpellName then
    return;
  end
  self.itemSpellName = itemSpellName;
  local validTarget = false;
  if UnitName(self.target)==target then
    validTarget = true;
  end
  if self.target=="target" and (target=="" or UnitName("target")==nil) then
    validTarget = true;
  end
  if validTarget then
    self.targetName = target;
    environment:RegisterEvent("UNIT_SPELLCAST_STOP", self);
    environment:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED", self);
    self.waitForCastResult = true;
  end  
end

function ItemAction:UNIT_SPELLCAST_STOP(unit)
  if unit=="player" then
    environment:UnregisterEvent("UNIT_SPELLCAST_STOP", self);      
    environment:UnregisterEvent("UNIT_SPELLCAST_SUCCEEDED", self);
    self.waitForCastResult = false;
  end
end

function ItemAction:UNIT_SPELLCAST_SUCCEEDED(unit, spellName)
  if unit=="player" then
    environment:UnregisterEvent("UNIT_SPELLCAST_STOP", self);      
    environment:UnregisterEvent("UNIT_SPELLCAST_SUCCEEDED", self);
    self.waitForCastResult = false;
    if spellName==self.itemSpellName then
      self:Fire("ACTION_PERFORMED", self, self.targetName);
    end
  end
end

function ItemAction:OnUsed(eventName)
  if eventName == "ACTION_AVAILABILITY_CHANGED" then
    environment:RegisterEvent("UNIT_INVENTORY_CHANGED", self);
    self.waitForAvailable = true;
  elseif eventName == "ACTION_PERFORMED" then
    environment:RegisterEvent("UNIT_SPELLCAST_SENT", self);      
  end
end

function ItemAction:OnUnused(eventName)
  if eventName == "ACTION_AVAILABILITY_CHANGED" then
    environment:UnregisterEvent("UNIT_INVENTORY_CHANGED", self);
    self.waitForAvailable = false;
  elseif eventName == "ACTION_PERFORMED" then
    if self.waitForCastResult then
      environment:UnregisterEvent("UNIT_SPELLCAST_STOP", self);      
      environment:UnregisterEvent("UNIT_SPELLCAST_SUCCEEDED", self);
      self.waitForCastResult = false;
    end
    environment:UnregisterEvent("UNIT_SPELLCAST_SENT", self);      
  end
end

function ItemAction:GetCooldown()
  return 0;
  -- FIXME : GetItemCooldown(self.itemLink);
end

function ItemAction:ShowInTooltip(tooltip)
  tooltip:SetHyperlink(self.itemLink);
end

function ItemAction:CanBeCast()
  return (IsUsableItem(self.value))
end

--[[
  The SpellActionFactory class.
--]]

-- spellName : string
function SpellActionFactory:Create(spellName, target)
  local spellAction = self:NewInstance();
  
  if target==nil then
    target = "target";
  end
  spellAction.target = target;

  spellAction.spellName = spellName;  
  spellAction.waitForCastResult = false;
  spellAction.waitForWeaponHand = false;
  spellAction.waitForDurationUpdate = false;
  spellAction.waitForAvailable = false;
  local name, rank, icon = GetSpellInfo(spellName);
  if name == nil then
    spellAction.available = false;
  else
    spellAction.available = false;
    spellAction.type = "spell";
    spellAction.value = name;
    if rank=="" then
      spellAction.shiftValue = name;    
    else
      spellAction.shiftValue = name..SpellAction.RANK_ONE;
    end
    spellAction.texture = icon;
  end
  return spellAction;
end

--[[
  The ItemActionFactory class.
--]]

function ItemActionFactory:Create(itemGID, target, equipNeeded)
  local itemAction = self:NewInstance();
  equipNeeded = true; -- do not manage item directly usable from bag

  if target==nil then
    target = "target";
  end
  
  itemAction.target = "target";

  itemAction.itemGID = itemGID;
  itemAction.waitForAvailable = false;
  itemAction.waitForCastResult = false;
  itemAction.available = false;
  itemAction.equipNeeded = equipNeeded;
  local name, itemLink, _, _, _, _, _, _, _, itemTexture = GetItemInfo(itemGID);
  
  if name == nil then
    itemAction.available = false;
  else
    if equipNeeded then
      itemAction.available = IsEquippedItem(itemGID);
    else
      itemAction.available = true;
    end
    itemAction.itemLink = itemLink; -- keep item link info for tooltip
    itemAction.type = "item";
    itemAction.value = name;
    itemAction.shiftValue = name;
    itemAction.texture = itemTexture;
  end
  return itemAction;
end

