--[[
    EventHandlerLib
    Copyright (C) 2008 Gael Lalire
    Contact: gael.lalire@laposte.net

    EventHandlerLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EventHandlerLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with EventHandlerLib.  If not, see <http://www.gnu.org/licenses/>.
--]]

local eventHandlerLib = LibStub:NewLibrary("EventHandlerLib-1.0", 1);
if not eventHandlerLib then
  return;
end

local factoryLib = LibStub("FactoryLib-1.0");

-- this is an util function
local function appendTable(table, i, size, ...)
  if i>size then
    return ...;
  end
  return table[i], appendTable(table, i+1, size, ...);
end

local EventHandler, EventHandlerFactory = factoryLib.Factory.CreateAbstractType();
eventHandlerLib.EventHandler = EventHandler;
eventHandlerLib.EventHandlerFactory = EventHandlerFactory;
EventHandler.inUseFunc = false;

-- do nothing (you may override it)
function EventHandler:OnUsed(eventName)
end

-- do nothing (you may override it)
function EventHandler:OnUnused(eventName)
end

function EventHandler:error(message)
  --[===[@debug@
  DEFAULT_CHAT_FRAME:AddMessage(debugstack());
  --@end-debug@]===]
  error(message);
end

function EventHandler:RegisterEvent(eventName, listener, ...)
  if listener==nil then
    self:error("EventHandler:RegisterEvent a listener could not be nil for event "..eventName);
  end
  local newUse = false;
  local listeners = self.listeners[eventName];
  if listeners == nil then
    newUse = true;
    listeners = {["size"] = 0, ["map"] = {}};
    self.listeners[eventName] = listeners;
  end
  local oldListener = listeners.map[listener];
  if oldListener == nil then
    -- add new listener
    listeners.size = listeners.size + 1;
    oldListener = {};
    listeners.map[listener] = oldListener;
  end
  oldListener.table = {...};
  oldListener.size = select("#", ...);
  -- do this at the end (so in OnUse fonction you can call Unregister and can't call Register again)
  if newUse then
    EventHandler.inUseFunc = true;
    self:OnUsed(eventName);
    EventHandler.inUseFunc = false;
  end
end

function EventHandler:UnregisterEvent(eventName, listener)
  local listeners = self.listeners[eventName];
  if listeners == nil then
    self:error("EventHandler:UnregisterEvent no listener registred for event "..eventName);
  end
  local notUse = false;
  local oldListener = listeners.map[listener];
  if oldListener == nil then
    self:error("EventHandler:UnregisterEvent this listener is not registred for event "..eventName);
  else
    -- remove a listener
    local nsize = listeners.size - 1;    
    if nsize == 0 then
      -- no more listener
      self.listeners[eventName] = nil;
      notUse = true;
    else
      listeners.size = nsize;
      listeners.map[listener] = nil;    
    end
  end
  -- do this at the end (so in OnUse fonction you can call Register and can't call Unregister again)
  if notUse then
    EventHandler.inUseFunc = true;
    self:OnUnused(eventName);
    EventHandler.inUseFunc = false;
  end
end

function EventHandler:Fire(eventName, ...)
  if EventHandler.inUseFunc then
    self:error("EventHandler:Fire it is not a good pratice to fire an event while OnUsed functions");
  end
  local listeners = self.listeners[eventName];
    if listeners ~= nil then
      local tmpMap = {};
      -- copy to tmpCall to prevent modification of listeners.map while calling funcs
      for k, v in pairs(listeners.map) do
        tmpMap[k] = v;
      end
      for k, v in pairs(tmpMap) do
        local func = k[eventName];
        if func==nil then
          self:error("EventHandler:Fire the handler does not define "..eventName.." function");
        else
          func(k, appendTable(v.table, 1, v.size, ...));
        end
      end
  end
end

function EventHandlerFactory:Create()
  eventHandler = self:NewInstance();
  eventHandler.listeners = {};
  return eventHandler;
end
