--[[
    TimerLib
    Copyright (C) 2008 Gael Lalire
    Contact: gael.lalire@laposte.net

    TimerLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    TimerLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with TimerLib.  If not, see <http://www.gnu.org/licenses/>.
--]]

local timerLib = LibStub:NewLibrary("TimerLib-1.0", 1);
if not timerLib then
  return;
end

-- begin import
local factoryLib = LibStub("FactoryLib-1.0");
local Factory = factoryLib.Factory;
-- end import

--[[
  The TimerRenderer class.
--]]

local TimerRenderer = Factory.CreateAbstractType();
timerLib.TimerRenderer = TimerRenderer;

-- this should be valid sentence :
-- - $name on $target is expiring
-- - $name sur $target va expirer
-- - so target should be the name of person concerned by timer (and not "player" which belongs to english langage)
-- if $target is "" then this should be valid sentence :
-- - $name is expiring
function TimerRenderer:StartTimer(name, duration, texture, target, count, tickFrequency, tickStart)
  error("TimerRenderer:StartTimer must be overrided");
end

function TimerRenderer:SetTimeLeft(timeLeft)
  error("TimerRenderer:SetTimeLeft must be overrided");
end

function TimerRenderer:StopTimer()
  error("TimerRenderer:StopTimer must be overrided");
end

function TimerRenderer:SetEnabled(enabled)
  error("TimerRenderer:SetEnabled must be overrided");
end

function TimerRenderer:SetCount(count)
  error("TimerRenderer:SetCount must be overrided");
end

--[[
  The Timer class.
--]]

local Timer, TimerFactory = Factory.CreateAbstractType();
timerLib.Timer = Timer;
timerLib.TimerFactory = TimerFactory;

function Timer:StartTimer(name, duration, texture, target, count, tickFrequency, tickStart, updater)
  if self.updater ~= nil then
    self.updater:Disable(self);
  end
  duration = duration or 0;
  -- if duration is 0 (or less) then we do not need to register updater and show timer renderers
  if duration > 0 then
    self.updater = updater;
    -- set default values
    name = name or "";
    target = target or "";
    count = count or -1;
    tickFrequency = tickFrequency or -1;
    tickStart = tickStart or -1;
    for i=1, #self.timerRendererList do
      self.timerRendererList[i]:StartTimer(name, duration, texture, target, count, tickFrequency, tickStart);
    end
    -- updater should be enabled after timer started
    if updater~=nil then
      updater:Enable(self);
    end
  else
    self.updater = nil;
    for i=1, #self.timerRendererList do
      self.timerRendererList[i]:StopTimer();
    end
  end
end

-- stop timer is a brutal event (buff dispell, totem destroy)
-- it is always call once by one of updater handler list (because we unregister after)
function Timer:StopTimer()
  for i=1, #self.timerRendererList do
    self.timerRendererList[i]:StopTimer();
  end
  if self.updater~=nil then
    self.updater:Disable(self);
    self.updater = nil;
  end
end

-- count is an integer between 0 and 99 (or -1)
-- can mean total person buffed, charge remaining ...
function Timer:SetCount(number)
  for i=1, #self.timerRendererList do
    self.timerRendererList[i]:SetCount(number);
  end
end

-- update left time ...
function Timer:SetTimeLeft(timeLeft)
  for i=1, #self.timerRendererList do
    self.timerRendererList[i]:SetTimeLeft(timeLeft);
  end
end

--[[
  The TimerFactory class.
--]]

function TimerFactory:Create(timerRendererList)
  local timer = self:NewInstance();
  timer.timerRendererList = timerRendererList;
  return timer;
end

--[[
  The ActionPerformedTimer class.
--]]

-- ActionPerformedTimer extends Timer
local ActionPerformedTimer, ActionPerformedTimerFactory, actionPerformedTimerFactory = Factory.CreateType(TimerFactory);
timerLib.ActionPerformedTimer = ActionPerformedTimer;
timerLib.ActionPerformedTimerFactory = ActionPerformedTimerFactory;
timerLib.actionPerformedTimerFactory = actionPerformedTimerFactory;

-- action performed do not send anything by default (maybe the action itself)
function ActionPerformedTimer:ACTION_PERFORMED(actionProperty, actionFunc, ...)
  -- actionFunc should start timer if needed
  actionFunc(self, actionProperty, ...);
end


--[[
  The ActionPerformedTimerFactory class.
--]]

function ActionPerformedTimerFactory:Create(timerRendererList, actionList, actionPropertyMap, actionFuncMap)
  local actionPerformedTimer = self:NewInstance(timerRendererList);
  
  if actionPropertyMap == nil then
    error("TimerFactory:CreateActionPerformedTimer actionPropertyMap cannot be nil");
  end
  if actionFuncMap == nil then
    actionFuncMap = {};
  end
  
  for i=1, #actionList do
    local action = actionList[i];
    action:RegisterEvent("ACTION_PERFORMED", actionPerformedTimer, actionPropertyMap[action], actionFuncMap[action]);
  end
  
  return actionPerformedTimer;
end

--[[
  The ActionCooldownTimer class.
--]]

-- ActionCooldownTimer extends Timer
local ActionCooldownTimer, ActionCooldownTimerFactory, actionCooldownTimerFactory = Factory.CreateType(TimerFactory);
timerLib.ActionCooldownTimer = ActionCooldownTimer;
timerLib.ActionCooldownTimerFactory = ActionCooldownTimerFactory;
timerLib.actionCooldownTimerFactory = actionCooldownTimerFactory;

-- action performed do not send anything by default (maybe the action itself)
function ActionCooldownTimer:COOLDOWN_UPDATED(actionProperty, actionFunc, ...)
  -- actionFunc should start timer if needed
  actionFunc(self, actionProperty, ...);
end

function ActionCooldownTimerFactory:Create(timerRendererList, action, actionProperty, actionFunc)
  local actionCooldownTimer = self:NewInstance(timerRendererList);
  actionCooldownTimer.action = action;
  action:RegisterEvent("COOLDOWN_UPDATED", actionCooldownTimer, actionProperty, actionFunc);
  return actionCooldownTimer;
end

