--[[
    TimerUpdaterLib
    Copyright (C) 2008 Gael Lalire
    Contact: gael.lalire@laposte.net

    TimerUpdaterLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    TimerUpdaterLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with TimerUpdaterLib.  If not, see <http://www.gnu.org/licenses/>.
--]]

local timerUpdaterLib = LibStub:NewLibrary("TimerUpdaterLib-1.0", 1);
if not timerUpdaterLib then
  return;
end

-- begin import
local factoryLib = LibStub("FactoryLib-1.0");
local Factory = factoryLib.Factory;
local environmentLib = LibStub("EnvironmentLib-1.0");
local environment = environmentLib.environment;
-- end import

--[[
  The TimerUpdater class.
--]]

TimerUpdater, TimerUpdaterFactory = Factory.CreateAbstractType();
timerUpdaterLib.TimerUpdater = TimerUpdater;
timerUpdaterLib.TimerUpdaterFactory = TimerUpdaterFactory;

-- take only environment event
function TimerUpdater:Enable(timer)
  for _, eventName in ipairs(self.environmentEventList) do
    if eventName=="UNIT_AURA" then
      environment:RegisterEventWithMinDelay(eventName, self, .8, timer);
    else
      environment:RegisterEvent(eventName, self, timer);
    end
  end
end

-- take only environment event
function TimerUpdater:Disable(timer)
  for _, eventName in ipairs(self.environmentEventList) do
    if eventName=="UNIT_AURA" then
      environment:UnregisterEventWithMinDelay(eventName, self);
    else
      environment:UnregisterEvent(eventName, self);
    end
  end
end

-- take only environment event
function TimerUpdaterFactory:Create(eventList)
  local timerUpdater = self:NewInstance();
  timerUpdater.environmentEventList = eventList or {};
  return timerUpdater;
end

--[[
  The WeaponTimerUpdater class.
--]]

local WeaponTimerUpdater, WeaponTimerUpdaterFactory, weaponTimerUpdaterFactory = Factory.CreateType(TimerUpdaterFactory);
timerUpdaterLib.WeaponTimerUpdater = WeaponTimerUpdater;
timerUpdaterLib.WeaponTimerUpdaterFactory = WeaponTimerUpdaterFactory;
timerUpdaterLib.weaponTimerUpdaterFactory = weaponTimerUpdaterFactory;

function WeaponTimerUpdater:Disable(timer)
  if self.waitForEnchantRemovedUpdate then
    environment:UnregisterEvent("UPDATE", self);
    self.waitForEnchantRemovedUpdate = false;
  end
  TimerUpdater.Disable(self, timer);
end

function WeaponTimerUpdater:COMBAT_LOG_EVENT_UNFILTERED(timer, _, event, _, _, _, _, unit, _, _, weapon)
  if unit ~= UnitName("player") then
    return;
  end
  if event == "ENCHANT_REMOVED" then
    local _, itemLink = GetItemInfo(weapon);
    local item = environment:GetItemGID(itemLink);
    local mainHandLink = environment:GetWeaponLink(true);
    local offHandLink = environment:GetWeaponLink(false);
    if self.mainHand then
      if mainHandLink ~= nil then
        if offHandLink == nil then
          timer:StopTimer();
        else
          -- here both main hand and offhand enchant
          local mainHand = environment:GetItemGID(mainHandLink);
          local offHand = environment:GetItemGID(offHandLink);
          if mainHand ~= offHand then
            if mainHand == item then
              timer:StopTimer();
            end
          else
            -- still unknown target ...
            local hasMainHandEnchant, _, _, hasOffHandEnchant = GetWeaponEnchantInfo();
            if hasMainHandEnchant and not hasOffHandEnchant then
              timer:StopTimer();
            else
              if hasOffHandEnchant then
                -- both weapons can be target ...
                self.waitForEnchantRemovedUpdate = true;
                environment:RegisterEvent("UPDATE", self, timer);
              end
            end
          end
        end
      end    
    else
      if offHandLink ~= nil then
        if mainHandLink == nil then
          timer:StopTimer();
        else
          -- here both main hand and offhand enchant
          local mainHand = environment:GetItemGID(mainHandLink);
          local offHand = environment:GetItemGID(offHandLink);
          if mainHand ~= offHand then
            if offHand == item then
              timer:StopTimer();
            end
          else
            -- still unknown target ...
            local hasMainHandEnchant, _, _, hasOffHandEnchant = GetWeaponEnchantInfo();
            if hasOffHandEnchant and not hasMainHandEnchant then
              timer:StopTimer();
            else
              if hasMainHandEnchant then
                -- both weapons can be target ...
                self.waitForEnchantRemovedUpdate = true;
                environment:RegisterEvent("UPDATE", self, timer);
              end
            end
          end
        end
      end    
    end
  elseif self.waitForEnchantRemovedUpdate and event == "ENCHANT_APPLIED"  then
    -- replacement of buff (can't check for remove)
    self.waitForEnchantRemovedUpdate = false;
    environment:UnregisterEvent("UPDATE", self);
  end
end

function WeaponTimerUpdater:UPDATE(timer)
  if not self.waitForEnchantRemovedUpdate then
    return;
  end

  local hasMainHandEnchant, _, _, hasOffHandEnchant = GetWeaponEnchantInfo();
  if self.mainHand then
    if not hasMainHandEnchant then
      self.waitForEnchantRemovedUpdate = false;
      environment:UnregisterEvent("UPDATE", self);
      timer:StopTimer();
    elseif not hasOffHandEnchant then
      self.waitForEnchantRemovedUpdate = false;
      environment:UnregisterEvent("UPDATE", self);      
    end
  else
    if not hasMainHandEnchant then
      self.waitForEnchantRemovedUpdate = false;
      environment:UnregisterEvent("UPDATE", self);
    elseif not hasOffHandEnchant then
      self.waitForEnchantRemovedUpdate = false;
      environment:UnregisterEvent("UPDATE", self);
      timer:StopTimer();
    end
  end
end

function WeaponTimerUpdater:UNIT_INVENTORY_CHANGED(timer)
  local hasMainHandEnchant, _, _, hasOffHandEnchant = GetWeaponEnchantInfo();
  if self.mainHand then
    if not hasMainHandEnchant or environment:GetWeaponLink(true) == nil then
      timer:StopTimer();
    end
  else
    if not hasOffHandEnchant or environment:GetWeaponLink(false) == nil then
      timer:StopTimer();
    end
  end
end

function WeaponTimerUpdaterFactory:Create(mainHand)
  local weaponTimerUpdater = self:NewInstance({"COMBAT_LOG_EVENT_UNFILTERED", "UNIT_INVENTORY_CHANGED"});
  weaponTimerUpdater.mainHand = mainHand;
  weaponTimerUpdater.waitForEnchantRemovedUpdate = false;
  return weaponTimerUpdater;
end

--[[
  The TotemTimerUpdater class.
--]]

local TotemTimerUpdater, TotemTimerUpdaterFactory, totemTimerUpdaterFactory = Factory.CreateType(TimerUpdaterFactory);
timerUpdaterLib.TotemTimerUpdater = TotemTimerUpdater;
timerUpdaterLib.TotemTimerUpdaterFactory = TotemTimerUpdaterFactory;
timerUpdaterLib.totemTimerUpdaterFactory = totemTimerUpdaterFactory;

function TotemTimerUpdater:PLAYER_TOTEM_UPDATE(timer)
  local _, _, _, duration = GetTotemInfo(self.slot);
  if duration<=0 then
    timer:StopTimer();
  end  
end

function TotemTimerUpdater:PLAYER_ENTERING_WORLD(timer)
  timer:StopTimer();
end

function TotemTimerUpdater:UNIT_AURA(timer)
  local totemBuffName = self.effect;
  if totemBuffName ~= nil then
    local count = 0;
    if environment:UnitBuff("player", totemBuffName) ~= nil then
      count = count + 1;
    end
    for i=1,4 do
      if environment:UnitBuff("party"..i, totemBuffName) ~= nil then
        count = count + 1;
      end
    end
    if self.raidEffect then
      for i=1,40 do
        if not UnitInParty("raid"..i) then
          if environment:UnitBuff("raid"..i, totemBuffName) ~= nil then
            count = count + 1;
          end
        end
      end    
    end
    timer:SetCount(count);  
  end
end

function TotemTimerUpdaterFactory:Create(slot)
  local totemTimerUpdater = self:NewInstance({"UNIT_AURA", "PLAYER_ENTERING_WORLD", "PLAYER_TOTEM_UPDATE"});
  totemTimerUpdater.slot = slot;
  return totemTimerUpdater;
end

--[[
  The BuffTimerUpdater class.
--]]

local BuffTimerUpdater, BuffTimerUpdaterFactory, buffTimerUpdaterFactory = Factory.CreateType(TimerUpdaterFactory);
timerUpdaterLib.BuffTimerUpdater = BuffTimerUpdater;
timerUpdaterLib.BuffTimerUpdaterFactory = BuffTimerUpdaterFactory;
timerUpdaterLib.buffTimerUpdaterFactory = buffTimerUpdaterFactory;


function BuffTimerUpdater:Enable(timer)
  TimerUpdater.Enable(self, timer);
  -- simulate a UNIT_AURA (because buff could happend before spell successful)
  self:UNIT_AURA(timer);
end

-- because bliz UNIT_AURA event sucks
function BuffTimerUpdater:UPDATE(timer)
  self:UNIT_AURA(timer);
end

function BuffTimerUpdater:UNIT_AURA(timer)
  local buff, _, _, count = environment:UnitBuff("player", self.buffName);  
  if not self.found then
    if buff ~= nil then
      self.found = true;
      timer:SetCount(count);  
    end
  else
    if buff == nil then
      timer:StopTimer();
    else
      timer:SetCount(count);    
    end
  end
end

function BuffTimerUpdaterFactory:Create(buffName)
  local buffTimerUpdater = self:NewInstance({"UNIT_AURA", "UPDATE"});
  buffTimerUpdater.buffName = buffName;
  buffTimerUpdater.found = false;
  return buffTimerUpdater;
end

--[[
  The FriendBuffTimerUpdater class.
--]]

local FriendBuffTimerUpdater, FriendBuffTimerUpdaterFactory, friendBuffTimerUpdaterFactory = Factory.CreateType(TimerUpdaterFactory);
timerUpdaterLib.FriendBuffTimerUpdater = FriendBuffTimerUpdater;
timerUpdaterLib.FriendBuffTimerUpdaterFactory = FriendBuffTimerUpdaterFactory;
timerUpdaterLib.friendBuffTimerUpdaterFactory = friendBuffTimerUpdaterFactory;

function FriendBuffTimerUpdater:UpdateTimer(timer)
  -- if the old path is not valid, refind path
  if self.targetPath==nil or environment:UnitFullName(self.targetPath)~=self.targetName then
    self.targetPath = environment:FindPath(self.targetName);
  end
  
  -- we lock the target of our earth shield
  if self.targetPath~=nil then
    local earthSieldBuff, _, _, count = environment:UnitBuff(self.targetPath, self.buffName);
    if count=="" then
      count = -1;
    else
      count = tostring(count);
    end
    if not self.found then
      if earthSieldBuff~= nil then
        self.found = true;
        self.timer = timer;
        timer:SetCount(count);      
      end
    else
      -- your target has not still your buff
      if earthSieldBuff == nil then
        timer:StopTimer();
      else
        timer:SetCount(count);      
      end
    end
  else
    timer:SetCount(-1);
  end
end

function FriendBuffTimerUpdater:UNIT_AURA(timer)
  self:UpdateTimer(timer);
end

function FriendBuffTimerUpdater:PLAYER_TARGET_CHANGED(timer)
  self:UpdateTimer(timer);
end

function FriendBuffTimerUpdater:PARTY_MEMBERS_CHANGED(timer)
  self:UpdateTimer(timer);
end


function FriendBuffTimerUpdaterFactory:Create(buffName)
  local buffTimerUpdater = self:NewInstance({"UNIT_AURA", "PLAYER_TARGET_CHANGED", "PARTY_MEMBERS_CHANGED"});
  buffTimerUpdater.buffName = buffName;
  buffTimerUpdater.found = false;
  return buffTimerUpdater;
end

--[[
  The FriendUniqueBuffTimerUpdater class.
--]]

local FriendUniqueBuffTimerUpdater, FriendUniqueBuffTimerUpdaterFactory, friendUniqueBuffTimerUpdaterFactory = Factory.CreateType(FriendBuffTimerUpdaterFactory);
timerUpdaterLib.FriendUniqueBuffTimerUpdater = FriendUniqueBuffTimerUpdater;
timerUpdaterLib.FriendUniqueBuffTimerUpdaterFactory = FriendUniqueBuffTimerUpdaterFactory;
timerUpdaterLib.friendUniqueBuffTimerUpdaterFactory = friendUniqueBuffTimerUpdaterFactory;

function FriendUniqueBuffTimerUpdater:Enable(timer)
  self.action:RegisterEvent("ACTION_PERFORMED", self, timer);
  -- super enable
  FriendBuffTimerUpdater.Enable(self, timer);
end

function FriendUniqueBuffTimerUpdater:Disable(timer)  
  -- super disable
  FriendBuffTimerUpdater.Disable(self, timer);
  self.action:UnregisterEvent("ACTION_PERFORMED", self);
end

function FriendUniqueBuffTimerUpdater:ACTION_PERFORMED(timer)
  timer:StopTimer();
end

function FriendUniqueBuffTimerUpdaterFactory:Create(buffName, action)
  local friendUniqueBuffTimerUpdater = self:NewInstance(buffName);
  friendUniqueBuffTimerUpdater.action = action;
  return friendUniqueBuffTimerUpdater;
end
