--[[
    LibSinkTimerRendererLib
    Copyright (C) 2008 Gael Lalire
    Contact: gael.lalire@laposte.net

    LibSinkTimerRendererLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    LibSinkTimerRendererLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with LibSinkTimerRendererLib.  If not, see <http://www.gnu.org/licenses/>.
--]]

local libSinkTimerRendererLib = LibStub:NewLibrary("LibSinkTimerRendererLib-1.0", 1);
if not libSinkTimerRendererLib then
  return;
end

-- begin import
local factoryLib = LibStub("FactoryLib-1.0");
local Factory = factoryLib.Factory;
local timerLib = LibStub("TimerLib-1.0");
local TimerRendererFactory = timerLib.TimerRendererFactory;
local environmentLib = LibStub("EnvironmentLib-1.0");
local environment = environmentLib.environment;
local localizationLib = LibStub("LocalizationLib-1.0");
local localization = localizationLib.localization;
-- end import

--[[
  The LibSinkTimerRenderer class.
--]]

-- LibSinkTimerRenderer extends TimerRenderer
local LibSinkTimerRenderer, LibSinkTimerRendererFactory, libSinkTimerRendererFactory = Factory.CreateType(TimerRendererFactory);
libSinkTimerRendererLib.LibSinkTimerRenderer = LibSinkTimerRenderer;
libSinkTimerRendererLib.LibSinkTimerRendererFactory = LibSinkTimerRendererFactory;
libSinkTimerRendererLib.libSinkTimerRendererFactory = libSinkTimerRendererFactory;

function LibSinkTimerRenderer:StartTimer(name, duration, texture, target, count, tickFrequency, tickStart)
  if not self.enabled then
    return;
  end
  self.name = name;
  if not self.started  then
    environment:RegisterEvent("UPDATE", self);
    self.started = true;
  else
    -- the timer is replaced with another (does not requires a message)
  end
  self.lastTimeLeft = duration;
  if duration > self.minDurationExpiring then
    self.testExpiring = true;
  else
    self.testExpiring = false;
  end
  self.lastTime = GetTime();
  self.lastTimeLeft = duration;
  self.lastCount = count;
  -- we do not care about tick (a message by tick is too many)  
  -- we could add texture when Pour but libSink does not allow it
  if self.activateStarted then
    self.libSink:Pour(localization:Format(self.startedMessage, self.name), 0, 10, 0);
  end
end

function LibSinkTimerRenderer:StopTimer()
  if self.started then
    environment:UnregisterEvent("UPDATE", self);
    if self.activateStopped then
      self.libSink:Pour(localization:Format(self.stoppedMessage, self.name), 10, 0, 0);
    end
    self.started = false;
  end
end

function LibSinkTimerRenderer:SetTimeLeft(timeLeft)
  if self.started then
    self.lastTimeLeft = GetTime() - self.lastTime + timeLeft;
  end
end

function LibSinkTimerRenderer:SetCount(count)
  if not self.enabled then
    return;
  end
  -- do not care about count change 
  if self.activateCountAtZero and count==0 and self.lastCount>0 then
    self.libSink:Pour(localization:Format(self.countAtZeroMessage, self.name), 10, 10, 0);
  end
  self.lastCount = count;
end

function LibSinkTimerRenderer:UPDATE()
  if not self.started then
    -- happend if StopTimer was called on same UPDATE event
    return;
  end
  local time = GetTime();
  local currentTimeLeft = self.lastTimeLeft - time + self.lastTime;
  if self.activateStopping and self.testExpiring then
    if self.lastTimeLeft>self.expiring and currentTimeLeft<=self.expiring then
      self.libSink:Pour(localization:Format(self.stoppingMessage, self.name), 10, 10, 0);
    end
  end
  if currentTimeLeft <= 0 then
    environment:UnregisterEvent("UPDATE", self);
    if self.activateStopped then
      self.libSink:Pour(localization:Format(self.stoppedMessage, self.name), 10, 0, 0);
    end
    self.started = false;
  end
  -- prepare next
  self.lastTimeLeft = currentTimeLeft;
  self.lastTime = time;
end

function LibSinkTimerRenderer:ActivatePour(activateStarted, activateStopping, activateStopped, activateCountAtZero)
  self.activateStarted = activateStarted or false;
  self.activateStopping = activateStopping or false;
  self.activateStopped = activateStopped or false;
  self.activateCountAtZero = activateCountAtZero or false;
end

function LibSinkTimerRenderer:SetEnabled(enabled)
  if enabled ~= self.enabled then
    self.enabled = enabled;
    if not enabled and self.started then
      environment:UnregisterEvent("UPDATE", self);
      self.started = false;
    end
  end
end

--[[
  The LibSinkTimerRendererFactory class.
--]]

function LibSinkTimerRendererFactory:Create(startedMessage, stoppingMessage, stoppedMessage, countAtZeroMessage,
 expiring, minDurationExpiring)
  libSinkTimerRenderer = self:NewInstance();
  libSinkTimerRenderer.startedMessage = startedMessage;
  libSinkTimerRenderer.stoppingMessage = stoppingMessage;
  libSinkTimerRenderer.stoppedMessage = stoppedMessage;
  libSinkTimerRenderer.countAtZeroMessage = countAtZeroMessage;
  libSinkTimerRenderer.activateStarted = false;
  libSinkTimerRenderer.activateStopping = false;
  libSinkTimerRenderer.activateStopped = false;
  libSinkTimerRenderer.activateCountAtZero = false;
  libSinkTimerRenderer.expiring = expiring or 5;
  libSinkTimerRenderer.minDurationExpiring = minDurationExpiring or 5;
  libSinkTimerRenderer.started = false;
  libSinkTimerRenderer.enabled = true;
  return libSinkTimerRenderer;
end

function LibSinkTimerRendererFactory:VARIABLES_LOADED()
  LibSinkTimerRenderer.libSink = {};
  setmetatable(LibSinkTimerRenderer.libSink, {__index = LibStub("LibSink-2.0")});
  environment:UnregisterEvent("VARIABLES_LOADED", self);
end

environment:RegisterEvent("VARIABLES_LOADED", libSinkTimerRendererFactory);
