--[[
    TM2TimerRendererLib
    Copyright (C) 2008 Gael Lalire
    Contact: gael.lalire@laposte.net

    TM2TimerRendererLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    TM2TimerRendererLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with TM2TimerRendererLib.  If not, see <http://www.gnu.org/licenses/>.
--]]

local tM2TimerRendererLib = LibStub:NewLibrary("TM2TimerRendererLib-1.0", 1);
if not tM2TimerRendererLib then
  return;
end

-- begin import
local factoryLib = LibStub("FactoryLib-1.0");
local Factory = factoryLib.Factory;
local timerLib = LibStub("TimerLib-1.0");
local TimerRendererFactory = timerLib.TimerRendererFactory;
local environmentLib = LibStub("EnvironmentLib-1.0");
local environment = environmentLib.environment;
-- end import


--[[
  The TM2TimerRenderer class.
--]]

-- TM2TimerRenderer extends TimerRenderer
local TM2TimerRenderer, TM2TimerRendererFactory, tM2TimerRendererFactory = Factory.CreateType(TimerRendererFactory);
tM2TimerRendererLib.TM2TimerRenderer = TM2TimerRenderer;
tM2TimerRendererLib.TM2TimerRendererFactory = TM2TimerRendererFactory;
tM2TimerRendererLib.tM2TimerRendererFactory = tM2TimerRendererFactory;


TM2TimerRenderer.LEFT_DIRECTION = 1;
TM2TimerRenderer.RIGHT_DIRECTION = 2;

function TM2TimerRenderer:StartTimer(name, duration, texture, target, count, tickFrequency, tickStart)
  self.name = name;
  if not self.started  then
    environment:RegisterEvent("UPDATE", self);
    self.started = true;
  end
    
  self.tickStart = tickStart;
  self.tickFrequency = tickFrequency;
  self.duration = duration;
  self.lastTimeLeft = duration;
  if duration > self.minDurationExpiring then
    self.testExpiring = true;
  else
    self.testExpiring = false;
  end
  self.lastTime = GetTime();
  self.lastTimeLeft = duration;
  self.lastCount = count;
  -- we do not care about tick (a message by tick is too many)  
  -- we could add texture when Pour but tM2 does not allow it
  
  -- do gui operations
  self.targetText:SetText(target);
  self:SetCount(count);
  self.framestatus:SetMinMaxValues(0, duration);
  self.texture:SetTexture(texture);
  self.frame:Show();
end

function TM2TimerRenderer:StopTimer()
  if self.started then
    environment:UnregisterEvent("UPDATE", self);
    self.frame:Hide();
    self.started = false;
  end
end

function TM2TimerRenderer:SetCount(count)
  -- do not care about count change 
  if count~=-1 then
    self.tics:SetText(count);
  else
    self.tics:SetText("");  
  end
end

function TM2TimerRenderer:SetTimeLeft(timeLeft)
  if self.started then
    self.lastTimeLeft = GetTime() - self.lastTime + timeLeft;
  end
end

function TM2TimerRenderer:UPDATE()
  if not self.started then
    -- happend if StopTimer was called on same UPDATE event
    return;
  end
  local time = GetTime();
  local currentTimeLeft = self.lastTimeLeft - time + self.lastTime;
  if self.activateStopping and self.testExpiring then
    if self.lastTimeLeft>self.expiring and currentTimeLeft<=self.expiring then
      -- later usage maybe
    end
  end
  if currentTimeLeft <= 0 then
    environment:UnregisterEvent("UPDATE", self);
    self.frame:Hide();
    self.started = false;
  end
  -- prepare next
  self.lastTimeLeft = currentTimeLeft;
  self.lastTime = time;
  
  -- do gui operations
  self.frameText:SetText(self:getTimeLeftText(currentTimeLeft));
  
  self.framestatus:SetValue(currentTimeLeft);
  
  if currentTimeLeft<=0 then
    self.ticstatus:SetValue(0);
  else
    if self.tickFrequency>0 then
      self.ticstatus:SetValue(1-((self.duration-currentTimeLeft-self.tickStart)%self.tickFrequency)/self.tickFrequency);
    else
      self.ticstatus:SetValue(1);
    end
  end
  
  
end

function TM2TimerRenderer:getTimeLeftText(tpsRestant)
  local secRestantes = floor(tpsRestant);
  if secRestantes>60 then
    return ceil(secRestantes/60).." m";
  else
    if tpsRestant>0 then return "|cffffffff"..secRestantes.." s"; -- affiche en blanc (peut retourner 0s)
    else return nil; end -- n'affiche rien a gerer
  end
end

function TM2TimerRenderer:SetDirection(direction)
  if direction==self.LEFT_DIRECTION then
    -- self.candyBar:SetCandyBarReversed(self.name, false);
  elseif direction==self.RIGHT_DIRECTION then
    -- self.candyBar:SetCandyBarReversed(self.name, true);
  end
end

function TM2TimerRenderer:SetPosition(parent, left, top)
  self.frame:SetPoint("LEFT", parent, "TOPLEFT", left, -top);
end

function TM2TimerRenderer:SetEnabled(enabled)
  if enabled and self.started then
    self.frame:Show();
  else
    self.frame:Hide();  
  end
end

--[[
  The TM2TimerRendererFactory class.
--]]

function TM2TimerRendererFactory:Create(parent, expiring, minDurationExpiring)
  tM2TimerRenderer = self:NewInstance();
  tM2TimerRenderer.activateStarted = false;
  tM2TimerRenderer.activateStopping = false;
  tM2TimerRenderer.activateStopped = false;
  tM2TimerRenderer.activateCountAtZero = false;
  tM2TimerRenderer.expiring = expiring or 5;
  tM2TimerRenderer.minDurationExpiring = minDurationExpiring or 5;
  tM2TimerRenderer.started = false;
    
  local frame = CreateFrame("Frame", nil, parent);
  tM2TimerRenderer.frame = frame;

  frame:SetWidth(100);
  frame:SetHeight(30);
  
  local bckd = {};
  bckd.bgFile = "Interface/Tooltips/UI-Tooltip-Background";
  bckd.edgeFile = "Interface/Tooltips/UI-Tooltip-Border";
  bckd.tile = true;
  bckd.tileSize = 16;
  bckd.edgeSize = 16;
  bckd.insets = {};
  bckd.insets.left = 4;
  bckd.insets.right = 4;
  bckd.insets.top = 4;
  bckd.insets.bottom = 4;

  frame:SetBackdrop(bckd);
  frame:SetBackdropColor(0, 0, 0, 1);
  frame:SetBackdropBorderColor(.3, .3, .3, 1);

  local texture = frame:CreateTexture();
  tM2TimerRenderer.texture = texture;
  texture:SetPoint("LEFT", frame, "LEFT", 5, 0);
  texture:SetWidth(20);
  texture:SetHeight(20);

  local tics = frame:CreateFontString(nil, "ARTWORK", "NumberFontNormal");
  tM2TimerRenderer.tics = tics;
  tics:SetAllPoints(texture);

  local ticstatus = CreateFrame("StatusBar", nil, frame);
  tM2TimerRenderer.ticstatus = ticstatus;
  ticstatus:SetPoint("TOPLEFT", frame, "TOPLEFT", 25, -5);

  ticstatus:SetStatusBarTexture("Interface\\TargetingFrame\\UI-StatusBar");
  ticstatus:SetStatusBarColor(0, 1, 0);
  ticstatus:SetMinMaxValues(0, 1);
  ticstatus:SetOrientation("VERTICAL");
  ticstatus:SetValue(1);
  ticstatus:SetWidth(5);
  ticstatus:SetHeight(20);
  
  local framestatus = CreateFrame("StatusBar", nil, frame);
  tM2TimerRenderer.framestatus = framestatus;
  
  framestatus:SetPoint("TOPLEFT", frame, "TOPLEFT", 30, -5);
  framestatus:SetStatusBarTexture("Interface\\TargetingFrame\\UI-StatusBar");
  framestatus:SetStatusBarColor(.5, .5, .5);
  framestatus:SetMinMaxValues(0, 1);
  framestatus:SetValue(0);
  framestatus:SetWidth(100-5*2-20-5);
  framestatus:SetHeight(20);

  local frameText = framestatus:CreateFontString(nil, "ARTWORK", "GameFontNormalSmall");
  tM2TimerRenderer.frameText = frameText;
  frameText:SetAllPoints();
  
  local targetText = frame:CreateFontString(nil, "ARTWORK", "GameFontNormalSmall");
  tM2TimerRenderer.targetText = targetText;
  targetText:SetPoint("TOP", frame, "BOTTOM");
  targetText:SetHeight(8);
  targetText:SetWidth(100);
  targetText:SetText("");
  
  frame:Hide();
    
  return tM2TimerRenderer;
end

